/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "config.h"
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <limits.h>
#include <fcntl.h>
#ifdef HAVE_LIBNCURSES
#include <curses.h>
#endif
#include <getopt.h>
#include <sys/types.h>
#include <sys/time.h>
#ifdef __CYGWIN__
#include <windows.h>
#undef ERROR
#endif
#include "version.h"
#include "main.h"
#include "error.h"
#include <sys/ioctl.h>

/************ Prototypes ************/

void main_trace_dumb(int n);
void main_trace_quiet(int);
void main_trace_curses(int n);
void main_finish_dumb();
void main_finish_quiet();
void main_finish_curses();
void main_init_curses();
void main_init_dumb();
void main_init_quiet();
void main_update_curses();
void main_update_dumb();
void main_update_quiet();
void main_do_options(int argc, char *argv[], char *envp[]);
void info();

/************************************/

char error_progname[PATH_MAX];
int main_paused = 0;

void main_trace_dumb(int n)
{
	static char src[PATH_MAX], dest[PATH_MAX], conv[PATH_MAX];
	static int isconv;
	int nr;

	if (n == 0) {
		isconv = 0;
		return;
	}
	nr = libgrab_getprogress(src,dest,conv);
	if (nr < 0) exit(0);
	if (main_paused) return;
	if (isconv) {
		if (!*src && !*dest && !*conv) return;
		printf("%s\n",conv);
	} else {
		if (!*src && !*dest && !*conv) {
			isconv = 1;
			return;
		}
		printf("%s\n",dest);
	}
}

void main_trace_quiet(int n)
{
	static char src[PATH_MAX], dest[PATH_MAX], conv[PATH_MAX];
	if (n == 0) return;
	if (libgrab_getprogress(src,dest,conv) < 0) exit(0);
}

#ifdef HAVE_LIBNCURSES
void main_trace_curses(int n)
{
	static int oldsec, oldusec, start, isconv;
	static double old2fps, oldfps, fps, bps, oldbps, old2bps, oldbits;
	static char src[PATH_MAX], dest[PATH_MAX], conv[PATH_MAX];
	static struct timeval after;
	int nr;
	extern int totalbits;

	if (n == 0) {
		gettimeofday(&after, 0);
		start = after.tv_sec;
		fps = oldfps = 0;
		bps = oldbps = 0;
		isconv = 0;
		return;
	}
	nr = libgrab_getprogress(src,dest,conv);
	if (nr < 0) exit(0);
	if (main_paused) return;
	if (isconv) {
		if (!*src && !*dest && !*conv) return;
		attron(A_BOLD);
		mvprintw(17,17,conv);
		attroff(A_BOLD);
		oldsec = after.tv_sec;
		oldusec = after.tv_usec;
		old2fps = oldfps;
		oldfps = fps;
		gettimeofday(&after, 0);
		fps = (1/((after.tv_sec + after.tv_usec/1.0e6) - (oldsec + oldusec/1.0e6)))+0.05;
		mvprintw(6, 16, "%2.01f (%2.01f)   ", (fps+oldfps+old2fps)/3,after.tv_sec==start?0:(double)nr/(double)(after.tv_sec-start));
		mvprintw(5, 47, "%i:%i.%i",(after.tv_sec-start)/3600,
			 ((after.tv_sec-start)/60)%60,(after.tv_sec-start)%60);
		if (totalbits) {
			old2bps = oldbps;
			oldbps = bps;
			bps = ((totalbits-oldbits)/((after.tv_sec + after.tv_usec/1.0e6) - (oldsec + oldusec/1.0e6)))+0.05;
			oldbits = totalbits;
			mvprintw(7, 16, "%2.01f (%2.01f) bps   ", (bps+oldbps+old2bps)/3,after.tv_sec==start?0:(double)totalbits/(double)(after.tv_sec-start));
		}
			
	} else {
		if (!*src && !*dest && !*conv) {
			isconv = 1;
			return;
		}
		attron(A_BOLD);
		mvprintw(17,17,dest);
		attroff(A_BOLD);
		oldsec = after.tv_sec;
		oldusec = after.tv_usec;
		old2fps = oldfps;
		oldfps = fps;
		gettimeofday(&after, 0);
		fps = (1/((after.tv_sec + after.tv_usec/1.0e6) - (oldsec + oldusec/1.0e6)))+0.05;
		mvprintw(6, 16, "%2.01f (%2.01f)   ", (fps+oldfps+old2fps)/3,after.tv_sec==start?0:(double)nr/(double)(after.tv_sec-start));
		mvprintw(5, 47, "%i:%i.%i",(after.tv_sec-start)/3600,
			 ((after.tv_sec-start)/60)%60,(after.tv_sec-start)%60);
	}

	refresh();
}

void main_finish_curses()
{
	flushinp();
	endwin();
}

void main_init_curses()
{
	initscr(); cbreak(); noecho();		// init ncurses
	leaveok(stdscr, TRUE);			// disable cursor
	wresize(stdscr,24,80);			// resize win to 80x24
	refresh();				// flush the screen 
	nodelay(stdscr,1);			// don't wait for input in getch
	start_color();				// set colors
	init_pair(1,COLOR_WHITE,COLOR_BLUE);
	init_pair(2,COLOR_RED,COLOR_BLUE);
	init_pair(3,COLOR_CYAN,COLOR_BLUE);
	wbkgd(stdscr, COLOR_PAIR(1));
	wattrset(stdscr, COLOR_PAIR(1));
	attron(A_BOLD);
	newwin(0,0,0,0);
	box(stdscr, 0, 0 );							// 80x24 box
	
	mvprintw(14, 2, "Keys : ");

	attroff(A_BOLD);
	mvprintw(14,10,"Q	- Quit");
	mvprintw(15,10,"P	- Pause");
	mvprintw(16,10,"S	- Sync HDDs");
	mvprintw(5, 32,"Time (secs.) :");
	mvprintw(6, 2, "FPS          :");
	mvprintw(9, 2, "Status       :");
	mvprintw(17, 2,"Writing      :");
	
	attron(A_BOLD);
	mvprintw(9, 17, "Working...");		// status

	wattrset(stdscr, COLOR_PAIR(2));
	mvprintw(3, 1, "------------------------------------------------------------------------------");
	mvprintw(11, 1, "------------------------------------------------------------------------------");
	wattrset(stdscr, COLOR_PAIR(3));
	attron(A_BOLD);
	mvprintw(0,2,"< bttvgrab " VERSION " (v4l) by J. Walter/A. Kopacz >");
	mvprintw(23,30,"< http://moes.pmnet.uni-oldenburg.de/bttvgrab/ >");
	wattrset(stdscr, COLOR_PAIR(1));
	
	refresh();
	
	wgetch(stdscr);
}

void main_update_curses()
{
	switch(getch()) { 
	case 'p': case 'P':
		attron(A_BOLD);
		if (!main_paused) {
			mvprintw(9,17,"PAUSED - Push any key (except s/q) to continue");
			libgrab_control(C_PAUSE);
			main_paused = 1;
		} else {
			mvprintw(9,17,"Working...                                    ");
			libgrab_control(C_CONT);
			main_paused = 0;
		}
		refresh();
		break;
	case 's': case 'S':
		attron(A_BOLD);
		mvprintw(10,17,"Syncing HDDs...");
		refresh();	      
		sync();
		attroff(A_BOLD);
		mvprintw(10, 17,"              ");
		refresh();
		break;
	case 'q': case 'Q':
		libgrab_control(C_STOP);
		exit(0);
		
	default: 
		break;
	}
}

void (*main_init_func)() = main_init_curses;
void (*main_trace_func)(int) = main_trace_curses;
void (*main_update)() = main_update_curses;
void (*main_finish_func)() = main_finish_curses;
#else
void (*main_init_func)() = main_init_dumb;
void (*main_trace_func)(int) = main_trace_dumb;
void (*main_finish_func)() = main_finish_dumb;
void (*main_update)() = main_update_dumb;
#endif

void main_finish_dumb()
{
#ifdef __CYGWIN__
	long cnt = 0;
	WriteFile(GetStdHandle(STD_OUTPUT_HANDLE),"done.\n",6,&cnt,0);
	FlushFileBuffers(GetStdHandle(STD_OUTPUT_HANDLE));
#else
	printf("done.\n");
#endif
}

void main_finish_quiet()
{
}

void main_finish()
{
#ifdef __CYGWIN__
	long cnt = 0;
#endif
	(*main_finish_func)();
	main_update = NULL;
	main_trace_func = NULL;
	if (*error_text) {
#ifdef __CYGWIN__
		strcat(error_text,"\n");
		WriteFile(GetStdHandle(STD_OUTPUT_HANDLE),error_text,strlen(error_text),&cnt,0);
		FlushFileBuffers(GetStdHandle(STD_OUTPUT_HANDLE));
#else
		fprintf(stderr,"%s\n",error_text);
#endif
	}
}

void main_init_dumb()
{
	setbuf(stdout,NULL);
}

void main_init_quiet()
{
	setbuf(stdout,NULL);
	setbuf(stdin,NULL);
}

void main_update_dumb()
{
	switch(getc(stdin)) { 
	case '\n':
		fflush(stdin);
		if (main_paused) {
			main_paused = 0;
			libgrab_control(C_CONT);
		} else {
			main_paused = 1;
			libgrab_control(C_PAUSE);
			printf("paused\n");
		}
		break;
	case 'q':
		fflush(stdin);
		libgrab_control(C_STOP);
		exit(0);
	default:
		break;
	}
}

void main_update_quiet()
{
}


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////                                             ///////////////
/////////////                 Main loop                   ///////////////
/////////////                                             ///////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[], char *envp[])
{
	fd_set fds;
	FD_ZERO(&fds);

	strcpy(error_progname,argv[0]);

	libgrab_init();
	libgrab_configure("input_file_fmt","pmm");
	libgrab_configure("grab_file_fmt","");
	main_do_options(argc, argv, envp);

	(*main_init_func)();
	(*main_trace_func)(0);
	atexit(main_finish);
	libgrab_control(C_START);
#ifndef NOTHREAD
loop:
	if (main_trace_func != main_trace_quiet) FD_SET(0,&fds);
	FD_SET(libgrab_errorpipe,&fds);
	FD_SET(libgrab_progresspipe,&fds);
	select(libgrab_errorpipe+libgrab_progresspipe+1,&fds,NULL,NULL,NULL);

	if (FD_ISSET(libgrab_errorpipe,&fds)) exit(1);
	if (FD_ISSET(libgrab_progresspipe,&fds) && main_trace_func) (*main_trace_func)(1);
	if (FD_ISSET(0,&fds) && main_update) (*main_update)();
	goto loop;
#endif
}
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

#define OPTSET(c,opt,val) case c: libgrab_configure(#opt,val); break;
#define OPT(c,opt) OPTSET(c,opt,optarg);

void main_do_options(int argc, char *argv[], char *envp[])
{
	int optchar, height = 0;
	char tmp[PATH_MAX], *p;

	while ((optchar = getopt(argc, argv, "P:S:I:O:N:m:W:w:x:b:B:rc:s:i:p:l:f:g:F:D:o:tyQd:hR:")) != -1) {
		switch (optchar) {	
		OPT('w',width);
		OPT('m',msleep);
		case 'W':
			height = 1;
		OPT(1,height);
		OPT('x',first_conv);
		OPT('b',video_bitrate);
		OPT('B',audio_bitrate);
		OPTSET('r',lowdisk,"1");
		OPT('c',conv_log);
		case 's':
			p = strchr(optarg,':');
			if (p) {
				*p = 0;
				libgrab_configure("input_transport",optarg);
				*p = ':';
				libgrab_configure("inputname",p+1);
			} else libgrab_configure("inputname",optarg);
			break;
		OPT('i',input_file_fmt);
		OPT('I',input_sound_fmt);
		OPT('O',output_sound_fmt);
		OPT('R',inputdata);
		OPT('p',outputdata);
		OPT('P',osounddata);
		OPT('l',limit);
		case 'f':
			p = strchr(optarg,':');
			if (p) {
				*p = 0;
				libgrab_configure("conv_transport",optarg);
				*p = ':';
				libgrab_configure("convname",p+1);
			} else libgrab_configure("convname",optarg);
			break;
		OPT('g',grablog);
		OPT('S',isoundname);
		OPT('F',osoundname);
		OPT('D',sounddevice);
		case 'o':
			libgrab_configure("conv_file_fmt",optarg);
			break;
		OPT('N',norm);
		OPTSET('t',do_trace,"1");
		OPTSET('y',do_sync,"1");
		OPTSET('Q',do_sound,"0");
		case 'd':
			ASSERT(strlen(optarg) == 1,(ARGPMSG("display type must be one of c,d,q")));
			switch (*optarg) {
			case 'c':
#ifdef HAVE_LIBNCURSES
				main_init_func = main_init_curses;
				main_trace_func = main_trace_curses;
				main_update = main_update_curses;
				main_finish_func = main_finish_curses;
#else
				ASSERT(0,(MSG("curses interface not compiled in")));
#endif
				break;
			case 'd':
				main_init_func = main_init_dumb;
				main_trace_func = main_trace_dumb;
				main_update = main_update_dumb;
				main_finish_func = main_finish_dumb;
				break;
			case 'q':
				main_init_func = main_init_quiet;
				main_trace_func = main_trace_quiet;
				main_update = main_update_quiet;
				main_finish_func = main_finish_quiet;
				break;
			default:
				ASSERT(0,(ARGPMSG("display type must be one of c,d,q")));
				break;
			}
			break;
		case 'h': 
		default:
			printf("bttvconvert " VERSION "\n");
			printf("(c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>\n");
			printf("Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/\n\n");

			printf("Options:\n");
			printf("-g [logname]     : file name of grab log, any options present in\n");
			printf("                   it will override preceding options\n");
			printf("                   if missing, you MUST specify -i and (for raw input) -w\n");
			printf("-t               : trace, allow gaps in file numbers as created by\n");
			printf("                   bttvgrab -t, you MUST specify -l as well\n");
			printf("-w [32-768]      : width, default = 640 (height=480)\n");
			printf("                   height is fixed at (3/4 * width)\n");
			printf("-o [format]      : output file format, default = first, available:\n");
			libgrab_inquire("conv_file_fmt list",tmp);
			printf("%s\n",tmp);
			printf("-i [format]      : input file format, default = pmm, available:\n");
			libgrab_inquire("input_file_fmt list",tmp);
			printf("%s\n",tmp);
			printf("-I [format]      : input sound format, default = first, available:\n");
			libgrab_inquire("input_sound_fmt list",tmp);
			printf("%s\n",tmp);
			printf("-O [format]      : output sound format, default = first, available:\n");
			libgrab_inquire("output_sound_fmt list",tmp);
			printf("%s\n",tmp);
			printf("-p [name]=[val]  : set extended output options, depends on output format\n");
			printf("                   try quality=n, many more available\n");
			printf("-P [name]=[val]  : set extended sound options, depends on output format\n");
			printf("-l [1-?]         : limit, exit after <limit> pictures.\n");
			printf("-s [URL]         : filename for input picture(s)\n");
			printf("-f [URL]         : filename for output picture(s)\n");
			printf("                   if it ends with '/', a default filename is appended\n");
			printf("                   otherwise use %%04d (or similar) for numbering\n");
			printf("                   currently supported methods: file: (default) and\n");
			printf("                   net: (use server address as name), and\n");
			printf("                   webcam: (like file:, plus some help files)\n");
			printf("-m [sleeptime]   : sleep time in msec between frames, useful for X output\n");
			printf("-d [c | d | q]   : display type (curses/dumb/quiet), default = curses\n");
			printf("-Q               : quiet, don't convert sound\n");
			printf("-S [filename]    : audio input filename/adress\n");
			printf("-F [filename]    : audio output filename/adress\n");
			printf("-h               : this help screen.\n\n");
			printf("\nMPEG options:\n");
			printf("-x [0-?]         : exclude first n frames, sound is synchronized accordingly\n");
			printf("-b [bitrate]     : video bitrate * 1000, default = 1152\n");
			printf("-B [bitrate]     : audio bitrate * 1000, default = 128\n");
			printf("-c [filename]    : conversion log, set this if you experience problems\n");
			printf("-r               : remove any used input files, DANGEROUS\n");
			printf("-N [normname]    : video norm, available: PAL, NTSC, SECAM\n");
			exit(0);
		}
	}
	if (!height) {
		libgrab_inquire("width",tmp);
		sprintf(tmp,"%i",atoi(tmp)*3/4);
		libgrab_configure("height",tmp);
	}
	libgrab_configure("grab_file_fmt","");
}
