/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef ERROR_H
#define ERROR_H
#include <errno.h>
#include <stdio.h>
extern char error_progname[];
extern char error_text[];
extern void error_abort(int);
extern int error_ignore;

#ifndef O_BINARY
#define O_BINARY 0
#endif

#ifdef ERROR_DEBUG
#include <stdarg.h>
#ifdef _WIN32
#include <windows.h>
#undef ERROR
#endif
static inline void error_printf (const char *fmt,...)
{
  char buf[4000];
  va_list ap;
#ifdef _WIN32
  DWORD done;
#endif
  int count;

  va_start (ap, fmt);
  count = vsprintf(buf, fmt, ap);
  va_end (ap);

#ifdef _WIN32
  WriteFile (GetStdHandle (STD_ERROR_HANDLE), buf, count, &done, 0);
  FlushFileBuffers (GetStdHandle (STD_ERROR_HANDLE));
#else
  printf("%s",buf);
  fflush(stdout);
#endif
}
#endif



#ifdef DEBUG
#undef DEBUG
#define DEBUG(x) "%s:%i:%s: " x ,__FILE__,__LINE__,__FUNCTION__
#else
#define DEBUG(x) x
#endif

#ifdef ASSERT
#undef ASSERT
#endif

#define ERROR                           if (!error_ignore) sprintf
#define ERROR_H_ERR(a)                  ERROR a
#define END(n)				do if (!error_ignore) { ENDFUNC(n); error_abort(n); } while (0)
#define ENDFUNC(n)

#define ERRMSG(a)                       error_text,DEBUG("%s error: %s (" a ")"),error_progname,strerror(errno)
#define MSG(a)                          error_text,DEBUG("%s error: " a),error_progname
#define NOERR(a)                        error_text,"%c",0
#define INTMSG(a)                       error_text,DEBUG("%s internal error: " a),error_progname
#define OPTMSG(a)			error_text,DEBUG("%s internal error: (%s.%s=%s) " a),error_progname,module,option,(set?NULL:value)
#define ARGMSG(a)                       error_text,DEBUG("%s error: invalid argument \"-%c\" - " a),error_progname,optchar
#define ARGPMSG(a)                      error_text,DEBUG("%s error: invalid argument \"-%c %s\" - " a),error_progname,optchar,optarg

#define ASSERT(a,b)                     do if (!(a)) { ERROR_H_ERR(b); END(1);} while (0)

#define NOTNULL(a,b)                    ASSERT((a) != NULL,b)
#define EQNULL(a,b)                     ASSERT((a) == NULL,b)
#define NOTNEG1(a,b)                    ASSERT((a) != -1,b)
#define RANGE(var,min,max,name,func)    ASSERT((var) >= (min) && (var) <= (max),(func(name " must be %i - %i"),min,max))
#define ARGRANGE(var,min,max,name)      RANGE(var,min,max,name,ARGPMSG)

#define OPENMSG(name)                   (ERRMSG("open %s"),name)
#define READMSG(name)                   (ERRMSG("read %s"),name)
#define WRITEMSG(name)                  (ERRMSG("write %s"),name)
#define CLOSEMSG(name)                  (ERRMSG("close %s"),name)

#define SAFE_REALLOC(var,size)          do { if (size) {NOTNULL((var) = realloc((var),(size)),(ERRMSG("realloc %i bytes"),(size)));} else {free(var); (var) = NULL;}} while (0)
#define SAFE_MALLOC(var,size)           NOTNULL((var) = malloc(size),(ERRMSG("malloc %i bytes"),(size)))
#define SAFE_CALLOC(var,n,size)         NOTNULL((var) = calloc((n),(size)),(ERRMSG("calloc %i bytes"),(n)*(size)))
#define SAFE_CREAT(var,name,mode)       NOTNEG1((var) = open((name),O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,(mode)),OPENMSG(name))
#define SAFE_OPEN(var,name,mode)        NOTNEG1((var) = open((name),(mode)|O_BINARY), OPENMSG(name))
#define SAFE_XOPEN(var,name,flags,mode) NOTNEG1((var) = open((name),(flags),(mode)|O_BINARY), OPENMSG(name))
#define SAFE_FOPEN(var,name,mode)       NOTNULL((var) = fopen((name),mode "b"),OPENMSG(name))
#define SAFE_CLOSE(var,name)            if (close(var) == -1 && errno != EBADF) ASSERT(0,CLOSEMSG(name))
#define SAFE_FCLOSE(var,name)           ASSERT(fclose(var) == 0,CLOSEMSG(name))
#define SAFE_WRITE(fh,buffer,size,name) ASSERT(write((fh),(buffer),(size)) == (size),WRITEMSG(name));
#define SAFE_FWRITE(buffer,size,fp,name) ASSERT(fwrite((buffer),01,(size),(fp)) == (size),WRITEMSG(name));
#define SAFE_READ(fh,buffer,size,name)  ASSERT(read((fh),(buffer),(size)) == (size),READMSG(name));
#define SAFE_FREAD(buffer,size,fp,name) ASSERT(fread((buffer),01,(size),(fp)) == (size),READMSG(name));
#define SAFE_FREE(var)                  do {if (var) {free(var); (var) = NULL;}} while (0)

#ifdef ERROR_DEBUG
#define CHK error_printf("%s:%i: checkpoint\n",__FILE__,__LINE__));
#endif

#endif
