/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* This file based on: */
/* TMN (H.263) coder version 3.0, University of British Columbia, CANADA, based on Telenor's coder version 2.0, Copyright (C) 1995, 1996 Telenor R&D, Norway */
/************************************************************************
 *
 *  main.c, main module of tmn (TMN encoder).
 *  tmn is an H.263+ (H.263 ver. 2.0) encoder somewhat based on the 
 *  Test Model Near-term (TMN8) in the ITU-T LBC Experts Group.
 *
 *  Copyright (C) 1997  University of BC, Canada
 *
 *  Contacts: 
 *  Michael Gallant                   <mikeg@ee.ubc.ca>
 *  Guy Cote                          <guyc@ee.ubc.ca>
 *  Berna Erol                        <bernae@ee.ubc.ca>
 *
 *  UBC Image Processing Laboratory   http://www.ee.ubc.ca/image
 *  2356 Main Mall                    tel.: +1 604 822 4051
 *  Vancouver BC Canada V6T1Z4        fax.: +1 604 822 5949
 *
 ************************************************************************/

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any
 * license fee or royalty on an "as is" basis. The University of
 * British Columbia disclaims any and all warranties, whether
 * express, implied, or statuary, including any implied warranties
 * or merchantability or of fitness for a particular purpose.  In no
 * event shall the copyright-holder be liable for any incidental,
 * punitive, or consequential damages of any kind whatsoever arising
 * from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs
 * and user's customers, employees, agents, transferees, successors,
 * and assigns.
 *
 * The University of British Columbia does not represent or warrant
 * that the programs furnished hereunder are free of infringement of
 * any third-party patents.
 *
 * Commercial implementations of H.263, including shareware, are
 * subject to royalty fees to patent holders.  Many of these patents
 * are general enough such that they are unavoidable regardless of
 * implementation design.
 *
*/

#include <math.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#define GLOBAL
#include "tmn/config.h"
#include "tmn/sim.h"
#include "error.h"
#include "configuration.h"
#include "transport.h"
#define SYMBOL(x) output_H263_ ## x

/************ Prototypes ************/

void output_H263_start(const char *unused);
int output_H263_write(unsigned char *buffer, int picnr);
void output_H263_stop();
void output_H263_DeterminePictureType(int *frame_no, Pict *pic, int P_skip, int B_skip);
void output_H263_CalculateStatistics(PictImage *curr_image, PictImage *curr_recon, PictImage *pb_b_image, PictImage *pb_b_recon, Bits *bits, int QP, Pict *pic);
int output_H263_FrameLayerRC(Pict *pic);

/************************************/

int output_H263_fmts() { return 1; }
const char *output_H263_fmt(int nr) { return "263"; }

float output_H263_mean_frame_rate, output_H263_ref_frame_rate, output_H263_frame_rate, output_H263_seconds;

int output_H263_frames,output_H263_Iframes,output_H263_bframes,output_H263_pframes,output_H263_Bframes, output_H263_EIframes, output_H263_EPframes;
int output_H263_wcopies,output_H263_icopies, output_H263_write_repeated;
int output_H263_total_frames_passed, output_H263_targetrate = 0;
int output_H263_startf, output_H263_end, output_H263_frame_no, output_H263_writediff;  
int output_H263_QP,output_H263_QPI;
int output_H263_first_frameskip, output_H263_chosen_frameskip = DEF_FRAMESKIP+1, output_H263_orig_frameskip, output_H263_frameskip;
int output_H263_mv_outside_frame_old, output_H263_source_format_old;
int output_H263_B_pictures_to_code;
int output_H263_prev_frame_no, output_H263_prev_I_P_frame_no, output_H263_prev_pict_type, output_H263_prev_I_P_pic_TR;
int output_H263_arith_used = 0;
int output_H263_start_rate_control = 0;
int output_H263_pb_in_progress = 0;

Bits output_H263_bits;
Bits output_H263_total_bits;
Results output_H263_total_res;
Bits output_H263_intra_bits;
Results output_H263_intra_res;
Bits output_H263_inter_bits;
Results output_H263_inter_res;
Bits output_H263_b_bits;
Results output_H263_b_res;
Bits output_H263_pb_bits;
Results output_H263_pb_res;
Bits output_H263_scal_bits;
Results output_H263_scal_res;
Bits output_H263_ei_bits;
Results output_H263_ei_res;
Bits output_H263_ep_bits;
Results output_H263_ep_res;

Results output_H263_res;

static int scal_tab[2][9] = { {1,1,1,2,2,1,1,2,2}, {1,1,1,1,1,2,2,2,2} };

PictImage *output_H263_prev_image = NULL;
PictImage *output_H263_curr_image = NULL;
PictImage *output_H263_curr_recon = NULL;
PictImage *output_H263_prev_recon = NULL;

  /* To be used for temporal scalability. */
PictImage *output_H263_next_P_image = NULL;
PictImage *output_H263_next_P_recon = NULL;
PictImage *output_H263_prev_P_image = NULL;
PictImage *output_H263_prev_P_recon = NULL;
  
  /* To be used for SNR/spatial scalability. */
PictImage *output_H263_prev_enhancement_image = NULL;
PictImage *output_H263_prev_enhancement_recon = NULL;
PictImage *output_H263_curr_reference_recon = NULL;
PictImage *output_H263_curr_enhancement_recon = NULL;
DiffImage *output_H263_diff_image = NULL;
DiffImage *output_H263_diff_recon= NULL;

  /* PB-frame specific */
PictImage *output_H263_B_recon = NULL;
PictImage *output_H263_B_image = NULL;

Pict output_H263_pic;
unsigned char *output_H263_image;

int output_H263_prev_I_P_quant, output_H263_first_P = 1;
int output_H263_True_B_frameskip, output_H263_P_frameskip;
int output_H263_TR_new = 0, output_H263_TR_old = 0;
int output_H263_rtype[MAX_LAYERS] = {0};

/* Off line rate control variables */
PictImage *output_H263_stored_image = NULL;

/* Reference Picture Selection variables */
int output_H263_thread_count = 0;
int output_H263_frame_in_thread = 0;
Thread ***output_H263_thread;

int output_H263_umv_mode = DEF_UMV_MODE;
int output_H263_reference_picture_selection = DEF_RPS_MODE;
int output_H263_BQUANT = DEF_BQUANT;
int output_H263_seek_dist = DEF_SEEK_DIST;
int output_H263_use_gobsync = DEF_INSERT_SYNC;
int output_H263_framerate = DEF_REF_FRAME_RATE;

int output_H263MV_fd = -1;
char output_H263_vectors[PATH_MAX] = "";

START_DATA
STROPTION(vectors);
INTALIAS(width,0,INT_MAX,pels);
INTALIAS(height,0,INT_MAX,lines);
INTOPTION(chosen_frameskip,1,INT_MAX);
CONSTINFO(color_fmt,"YUV420P");
BOOLOPTION(adv_pred);
BOOLOPTION(syntax_arith_coding);
BOOLOPTION(umv_mode);
BOOLOPTION(pb_frames);
BOOLOPTION(improved_pb_frames);
BOOLOPTION(advanced_intra_coding);
BOOLOPTION(deblocking_filter);
BOOLOPTION(slice_structure);
BOOLOPTION(reference_picture_selection);
INTOPTION(number_of_threads,1,INT_MAX);
INTOPTION(frames_per_thread,1,INT_MAX);
INTOPTION(sync_gobs_per_frame,1,INT_MAX);
BOOLOPTION(independent_segment_decoding);
BOOLOPTION(alternative_inter_vlc);
BOOLOPTION(modified_quantization);
INTOPTION(scalability_mode,0,7);
BOOLOPTION(reference_picture_resampling);
BOOLOPTION(reduced_resolution_update);
INTOPTION(successive_B_frames,0,INT_MAX);
INTOPTION(b_picture_quant,1,31);
INTOPTION(enhancement_layer_quant,1,31);
BOOLOPTION(unlimited_unrestricted_motion_vectors);
INTOPTION(force_intra,0,INT_MAX);
INTOPTION(intra_mb_refresh,0,INT_MAX);
INTOPTION(BQUANT,0,3);
INTOPTION(rate_control_method,0,3);
INTOPTION(start_rate_control,0,100);
INTOPTION(seek_dist,0,INT_MAX);
INTOPTION(use_gobsync,0,INT_MAX);
INTOPTION(p_picture_quant,1,31);
INTOPTION(i_picture_quant,1,31);
INTOPTION(framerate,1,60);
INTOPTION(targetrate,0,INT_MAX);
END_DATA

#define MVBUF_MAX 400
#define VEC_MAX 32
short output_H263MV_vec[VEC_MAX][2];
int output_H263MV_vcur;
short output_H263MV_buf[MVBUF_MAX][2];
int output_H263MV_bcur;
int output_H263MV_last;

void output_H263_start(const char *unused)
{
	int i,j;
	output_H263_pic.BCM = (BCM *)calloc(1,sizeof(BCM));

#ifndef FASTIDCT
	output_H263_init_idctref();
#endif

	output_H263_pic.unrestricted_mv_mode = output_H263_umv_mode;
	output_H263_mv_outside_frame = output_H263_umv_mode || output_H263_adv_pred || output_H263_deblocking_filter;
	output_H263_long_vectors = output_H263_umv_mode;
	output_H263_B_pictures_to_code = 0;
	output_H263_use_4mv = output_H263_adv_pred || output_H263_deblocking_filter;
	output_H263_overlapping_MC = output_H263_adv_pred;

	if (output_H263_improved_pb_frames || output_H263_advanced_intra_coding || output_H263_deblocking_filter
	    || output_H263_slice_structure || output_H263_pic.reference_picture_selection || output_H263_independent_segment_decoding
	    || output_H263_alternative_inter_vlc || output_H263_modified_quantization || output_H263_scalability_mode
	    || output_H263_reference_picture_resampling || output_H263_reduced_resolution_update || output_H263_successive_B_frames) {
		output_H263_EPTYPE = 1;
	} else {
		output_H263_EPTYPE = 0;
	}

  
	output_H263_pic.RLNUM = BASE_LAYER_NUM;
	output_H263_pic.ELNUM = BASE_LAYER_NUM;

	output_H263_pic.BQUANT = output_H263_BQUANT;

	output_H263_pic.seek_dist = output_H263_seek_dist;
	output_H263_pic.use_gobsync = output_H263_use_gobsync;
	output_H263_pic.sync = 0;

	output_H263_ref_frame_rate = (float)output_H263_framerate;
	output_H263_orig_frameskip = 1;

	output_H263_mv_outside_frame_old = output_H263_mv_outside_frame;
	output_H263_prev_I_P_pic_TR = output_H263_startf = DEF_START_FRAME;
	output_H263_end = DEF_STOP_FRAME;
  
	output_H263_frames = 0;
	output_H263_pframes = 0;
	output_H263_bframes = 0;
	output_H263_Bframes = 0;
	output_H263_total_frames_passed = 0;
	output_H263_pic.PB = 0;

	output_H263_pic.TR = 0;
	output_H263_pic.QP_mean = (float)0.0;

	output_H263_pic.UFEP = 1;
	output_H263_pic.PCF = 0;
	output_H263_pic.TRPI = 0;
	output_H263_pic.PEI = 0;
	output_H263_pic.MODB =0;
	output_H263_pic.GFID = 0;

	/* Initialize random generator */
	srand(time(NULL));

	if (output_H263_pels == 128 && output_H263_lines == 96) output_H263_pic.source_format = (SF_SQCIF);
	else if (output_H263_pels == 176 && output_H263_lines == 144) output_H263_pic.source_format = (SF_QCIF);
	else if (output_H263_pels == 352 && output_H263_lines == 288) output_H263_pic.source_format = (SF_CIF);
	else if (output_H263_pels == 704 && output_H263_lines == 576) output_H263_pic.source_format = (SF_4CIF);
	else if (output_H263_pels == 1408 && output_H263_lines == 1152) output_H263_pic.source_format = (SF_16CIF);
	else if (!(output_H263_pels%16 || output_H263_pels%16)) output_H263_pic.source_format = (SF_CUSTOM);
	else {
		ERROR(MSG("invalid image size for H.263, only 128x96, 176x144, 352x288, 704x576, or multiples of 16 are available"));
		END(1);
	}
	output_H263_cpels = output_H263_pels/2;
	output_H263_base_pels = output_H263_pels;
	output_H263_base_lines = output_H263_lines;
	output_H263_enhancement_pels  = output_H263_pels*scal_tab[0][output_H263_scalability_mode];
	output_H263_enhancement_lines = output_H263_lines*scal_tab[1][output_H263_scalability_mode];

	if (output_H263_successive_B_frames >= output_H263_chosen_frameskip) {
		output_H263_successive_B_frames = output_H263_chosen_frameskip - 1;
	}

	ASSERT(!(output_H263_successive_B_frames && (output_H263_pb_frames || output_H263_improved_pb_frames)),(MSG("Encoder supports either PB frames or true B pictures, not both.")));
	ASSERT(!(output_H263_successive_B_frames && output_H263_scalability_mode),(MSG("Encoder does not yet support multiple scalability options.")));
	ASSERT(!(output_H263_syntax_arith_coding && (output_H263_alternative_inter_vlc || output_H263_modified_quantization)),(MSG("Syntax based Arithmetic Coding cannot be used with Alternative Inter VLC, Improved PB or  Modified Quantization modes.")));
	ASSERT(!(output_H263_EPTYPE==ON && output_H263_pb_frames),(MSG("PB frames (annex G) mode cannot be use with any of the H.263 PLUS options. Use Improved PB frames (annex M) mode instead.")));
	output_H263_pic.target_frame_rate = (output_H263_ref_frame_rate / (output_H263_orig_frameskip * output_H263_chosen_frameskip));

	/* Allocate memory for thread structure */
	if (output_H263_pic.reference_picture_selection) {
		output_H263_thread = (Thread ***) malloc(output_H263_number_of_threads*sizeof(Thread));
		for (j=0; j<output_H263_number_of_threads; ++j) {
			output_H263_thread[j] = (Thread **)malloc(output_H263_frames_per_thread*sizeof(Thread) );
			for (i=0; i<output_H263_frames_per_thread; ++i) {
				output_H263_thread[j][i] = (Thread *)calloc(1,sizeof(Thread) );
			}
		} 
	}

	/* rate control variables */
	if (output_H263_targetrate != 0) {
		output_H263_pic.bit_rate = output_H263_targetrate;
		output_H263_InitializeRateControl();
	}
	
	/* pic->src_frame_rate needed to update UFEP */
	output_H263_pic.src_frame_rate = (int)(output_H263_ref_frame_rate / output_H263_orig_frameskip);

	output_H263_frame_rate =  output_H263_ref_frame_rate / (float)(output_H263_orig_frameskip * output_H263_chosen_frameskip);

	/* Initialize bitcounters */
	output_H263_initbits();

	/* Needed for Annex O temporal scalability 
	 * fixed_frameskip = chosen_frameskip x orig_frameskip; 
	 * True_B_frameskip = fixed_frameskip; */
	output_H263_True_B_frameskip = output_H263_frameskip / (output_H263_successive_B_frames + 1);

	/* number of seconds to encode */
	output_H263_seconds = (output_H263_end - output_H263_startf + output_H263_chosen_frameskip) * output_H263_orig_frameskip/ output_H263_ref_frame_rate;
  
	output_H263_pic.UFEP = 0;
	output_H263_prev_I_P_frame_no = output_H263_startf;
	output_H263_pb_in_progress = 0;

	transport_wstart("263",output_H263_pels,output_H263_lines,0);

	/* Intra */
	output_H263_pic.picture_coding_type = PCT_INTRA;

	output_H263_frame_no = output_H263_prev_I_P_frame_no = output_H263_startf;
	output_H263_prev_I_P_pic_TR = output_H263_pic.TR = output_H263_prev_I_P_pic_TR;

#ifdef VISI
	output_H263MV_vcur = 1;
	memset(output_H263MV_vec,0,sizeof(output_H263MV_vec));
	output_H263MV_vec[0][0] = output_H263_pels;
	output_H263MV_vec[0][1] = output_H263_lines;
	output_H263MV_last = -1;
#endif
}

int output_H263_picnr;
int output_H263_write(unsigned char *buffer, int picnr)
{
	int j;

#ifdef VISI
	output_H263MV_bcur = 0;
	memset(output_H263MV_buf,0,sizeof(output_H263MV_buf));
#endif
	transport_wframe(picnr);
	output_H263_picnr = picnr;

	if (output_H263_pb_in_progress) {
		output_H263_pic.picture_coding_type = PCT_PB_CONT;
	} else {
		if (picnr < output_H263_frame_no) return 0;
		else output_H263_frame_no = picnr;
	}

	switch (output_H263_pic.picture_coding_type) {
        case PCT_INTRA:
		++output_H263_Iframes;
		output_H263_pels = output_H263_base_pels;
		output_H263_lines = output_H263_base_lines;
		output_H263_cpels = output_H263_base_pels/2;

		output_H263_curr_image = output_H263_FillImage(buffer);
		output_H263_pic.QUANT = output_H263_QP = output_H263_i_picture_quant;
		output_H263_pic.DQUANT = 0;
		if (output_H263_EPTYPE) {
			/* Alternate picture rounding type */
			output_H263_pic.RTYPE = output_H263_rtype[0] % 2;
			output_H263_rtype[0]++;
		}
		output_H263_curr_recon = output_H263_CodeOneIntra(output_H263_curr_image, output_H263_QP, &output_H263_bits, &output_H263_pic);
		output_H263_CalculateStatistics(output_H263_curr_image, output_H263_curr_recon, NULL, NULL, &output_H263_bits, output_H263_QP, &output_H263_pic);
		/* Special handling if temporal scalability. B frames will be 
		 * inserted between the first I-P pair. So the future P frame will
		 * be the next frame coded. It sets prev_P_image and prev_P_recon
		 * based on to the previous next_P_image and next_P_recon, 
		 * respectively. */
		if (output_H263_successive_B_frames) {
			/* Free the past prev_P_image and prev_P_recon, no longer needed */
			if (NULL != output_H263_prev_P_image) {
				output_H263_FreeImage(output_H263_prev_P_image);
				output_H263_FreeImage(output_H263_prev_P_recon);
			}
			/* Set new prev_P_image and prev_P_recon to old next_P_image and 
			 * next_P_recon respectively */
			output_H263_prev_P_image = output_H263_next_P_image;
			output_H263_prev_P_recon = output_H263_next_P_recon;

			/* For future I_P pictures, curr_image and curr_recon are set to 
			 * prev_image and prev_recon. For B picture prediction, curr_image
			 * and curr_recon represent next_P_image and next_P_recon. */
			output_H263_prev_image = output_H263_next_P_image = output_H263_curr_image;
			output_H263_prev_recon = output_H263_next_P_recon = output_H263_curr_recon;
		} else {
			if (NULL != output_H263_prev_image) {
				output_H263_FreeImage(output_H263_prev_image);
				output_H263_FreeImage(output_H263_prev_recon);  
			}
			output_H263_prev_image = output_H263_curr_image;
			output_H263_prev_recon = output_H263_curr_recon;
		}

		output_H263_prev_I_P_quant = output_H263_pic.QUANT;
        
		/* Frame layer Rate Control */
		output_H263_frameskip = output_H263_FrameLayerRC(&output_H263_pic);

		break;

	case PCT_INTER:
		
		output_H263_pframes++;
		output_H263_frames++;
        
		output_H263_pels = output_H263_base_pels;
		output_H263_lines = output_H263_base_lines;
		output_H263_cpels = output_H263_base_pels/2;
        
		output_H263_curr_image = output_H263_FillImage(buffer);
               
		/* Set QP to output_H263_pic.QUANT from previous encoded picture */
		if (output_H263_first_P) {
			output_H263_first_P = 0;
			output_H263_prev_I_P_quant = output_H263_p_picture_quant;
		}

		output_H263_QP = output_H263_prev_I_P_quant;
		output_H263_pic.DQUANT = 0;
		if (output_H263_EPTYPE) {
			/* Alternate picture rounding type */
			output_H263_pic.RTYPE = output_H263_rtype[0] % 2;
			output_H263_rtype[0]++;
		}
		output_H263_curr_recon = output_H263_InitImage(output_H263_pels*output_H263_lines);

		output_H263_CodeOneOrTwo(output_H263_curr_image, NULL, output_H263_prev_image, output_H263_prev_recon, output_H263_QP, output_H263_P_frameskip * output_H263_orig_frameskip, &output_H263_bits, &output_H263_pic, NULL, output_H263_curr_recon);

		output_H263_CalculateStatistics(output_H263_curr_image, output_H263_curr_recon, NULL, NULL, &output_H263_bits, output_H263_QP, &output_H263_pic);

		/* Special handling for case of temporal scalability. The P frame being coded
		 * is the future P frame. */
		if (output_H263_successive_B_frames) {
			/* Free the past prev_P_image and prev_P_recon, no longer needed */
			if (NULL != output_H263_prev_P_image) {
				output_H263_FreeImage(output_H263_prev_P_image);
				output_H263_FreeImage(output_H263_prev_P_recon);
			}

			/* Set new prev_P_image and prev_P_recon to old next_P_image and 
			 * next_P_recon respectively */
			output_H263_prev_P_image = output_H263_next_P_image;
			output_H263_prev_P_recon = output_H263_next_P_recon;

			/* For future I_P pictures, curr_image and curr_recon are set to 
			 * prev_image and prev_recon. For B picture prediction, curr_image
			 * and curr_recon represent next_P_image and next_P_recon. */
			output_H263_prev_image = output_H263_next_P_image = output_H263_curr_image;
			output_H263_prev_recon = output_H263_next_P_recon = output_H263_curr_recon;
		} else {
			if (NULL != output_H263_prev_recon) {
				output_H263_FreeImage(output_H263_prev_recon);
				if ( (!output_H263_scalability_mode) || (output_H263_scalability_mode && (output_H263_prev_image != output_H263_prev_enhancement_image) ) ) {
					output_H263_FreeImage(output_H263_prev_image);
				}
			}
			output_H263_prev_image = output_H263_curr_image;
			output_H263_prev_recon = output_H263_curr_recon;
		}

		/* Frame Layer Rate Control */
		output_H263_frameskip = output_H263_FrameLayerRC(&output_H263_pic);
		output_H263_prev_I_P_quant = output_H263_pic.QUANT;

		/* Update frameskip for true B frames */
		output_H263_True_B_frameskip = output_H263_P_frameskip / (output_H263_successive_B_frames + 1);

		break;

	case PCT_PB:
	case PCT_IPB:
		output_H263_bframes++;
		output_H263_pframes++;
		output_H263_frames += 2;
		
		output_H263_pels = output_H263_base_pels;
		output_H263_lines = output_H263_base_lines;
		output_H263_cpels = output_H263_base_pels/2;
		
		/* Set QP to output_H263_pic.QUANT from previous encoded picture */
		output_H263_QP = output_H263_prev_I_P_quant;
		output_H263_pic.DQUANT = 0;
		if (output_H263_EPTYPE) {
			/* Alternate picture rounding type */
			output_H263_pic.RTYPE = output_H263_rtype[0] % 2;
			output_H263_rtype[0]++;
		}
//		image = ReadImage(seqfilename,frame_no,headerlength);
		output_H263_curr_image = output_H263_FillImage(buffer);
		output_H263_B_image = output_H263_curr_image;
		output_H263_frame_no += output_H263_P_frameskip;
		output_H263_pb_in_progress = 1;
		return 1;

	case PCT_PB_CONT:
//		image = ReadImage(seqfilename,frame_no + P_frameskip,headerlength);
		output_H263_pb_in_progress = 0;
		output_H263_curr_image = output_H263_FillImage(buffer);

		output_H263_pic.TRB = output_H263_P_frameskip * output_H263_orig_frameskip;
		if (output_H263_pic.TRB>7) output_H263_pic.TRB=7;
      
		/* Code two frames as a PB-frame */
		output_H263_B_recon = output_H263_InitImage(output_H263_pels*output_H263_lines);
		output_H263_curr_recon = output_H263_InitImage(output_H263_pels*output_H263_lines);
		
		/* Code PP or PB pictures. */
		output_H263_CodeOneOrTwo(output_H263_curr_image, output_H263_B_image, output_H263_prev_image, output_H263_prev_recon, output_H263_QP, 2*output_H263_P_frameskip*output_H263_orig_frameskip, &output_H263_bits, &output_H263_pic, output_H263_B_recon, output_H263_curr_recon);

		if (NULL != output_H263_prev_recon) {
			output_H263_FreeImage(output_H263_prev_recon);
			if ( (!output_H263_scalability_mode) || (output_H263_scalability_mode && (output_H263_prev_image != output_H263_prev_enhancement_image) ) ) {
				output_H263_FreeImage(output_H263_prev_image);
			}
		}

		output_H263_prev_image = output_H263_curr_image;
		output_H263_prev_recon = output_H263_curr_recon;

		output_H263_CalculateStatistics(output_H263_curr_image, output_H263_curr_recon, output_H263_B_image, output_H263_B_recon,  &output_H263_bits, output_H263_pic.BQUANT, &output_H263_pic);

		/* Frame Layer Rate Control */
		output_H263_frameskip = output_H263_FrameLayerRC(&output_H263_pic);
        
		output_H263_prev_I_P_quant = output_H263_pic.QUANT;

		break;

	case PCT_B:
		output_H263_Bframes++;
        
		output_H263_pels = output_H263_base_pels;
		output_H263_lines = output_H263_base_lines;
		output_H263_cpels = output_H263_base_pels/2;

		output_H263_curr_image = output_H263_FillImage(buffer);
        
		/* Set QP to b_picture_quant, no rate control for true B yet */
		output_H263_QP = output_H263_b_picture_quant;
		output_H263_pic.DQUANT = 0;
		output_H263_rtype[0]++;
  
		/* Code the next frame as true-B. */
		output_H263_curr_recon = output_H263_InitImage(output_H263_pels*output_H263_lines);

		output_H263_mv_outside_frame_old = output_H263_mv_outside_frame;
		output_H263_mv_outside_frame = ON;

		/* Code the current B picture. */
		output_H263_CodeOneTrueB(output_H263_next_P_image, output_H263_curr_image, output_H263_prev_P_image, output_H263_prev_P_recon, output_H263_QP, output_H263_True_B_frameskip * output_H263_orig_frameskip, &output_H263_bits, &output_H263_pic, output_H263_curr_recon, output_H263_next_P_recon);

		output_H263_CalculateStatistics(output_H263_curr_image, output_H263_curr_recon, NULL, NULL, &output_H263_bits, output_H263_QP, &output_H263_pic);

		output_H263_mv_outside_frame = output_H263_mv_outside_frame_old;

		break;

	case PCT_EI:

		output_H263_EIframes++;
      
		output_H263_pels = output_H263_enhancement_pels;
		output_H263_lines = output_H263_enhancement_lines;
		output_H263_cpels = output_H263_enhancement_pels/2;
		
		/* QP fixed for enhancement layer */
		output_H263_QP = output_H263_enhancement_layer_quant;
		output_H263_pic.DQUANT = 0;

		output_H263_pic.UFEP = 1;
		output_H263_pic.RLNUM = BASE_LAYER_NUM;
		output_H263_pic.ELNUM = ENHANCEMENT_LAYER_NUM;
       
		/* Alternate picture rounding type */
		output_H263_pic.RTYPE = output_H263_rtype[output_H263_pic.ELNUM-1] % 2;
		output_H263_rtype[output_H263_pic.ELNUM-1]++;
        
		/* For all possible spatial scalings */
		if (output_H263_scalability_mode >= 3) {
			/* No MVs from reference layer, so don't need to frame recon */
			output_H263_curr_reference_recon = output_H263_UpsampleReferenceLayerPicture(output_H263_prev_recon);

			/* Enhancement layer image is temporally simultaneous, same image as was
			 * just coded in reference layer, i.e. stored as prev_image */
			output_H263_curr_image = output_H263_UpsampleReferenceLayerPicture(output_H263_prev_image);
			
			output_H263_source_format_old = output_H263_pic.source_format;
			output_H263_pic.source_format = (output_H263_scalability_mode>=7) ? output_H263_source_format_old+1 : SF_CUSTOM;
		} else {
			output_H263_curr_reference_recon = output_H263_prev_recon;
			output_H263_curr_image = output_H263_prev_image;
		}

		output_H263_mv_outside_frame_old = output_H263_mv_outside_frame;
		output_H263_mv_outside_frame = ON;
 
		output_H263_curr_enhancement_recon = output_H263_InitImage(output_H263_pels*output_H263_lines);
        
		output_H263_CodeOneEI(output_H263_curr_image, output_H263_curr_reference_recon, output_H263_QP, &output_H263_bits, &output_H263_pic, output_H263_curr_enhancement_recon);

		/* Must apply DF here, as only now do we have fully reconstructed, enhanced picture */
		if (output_H263_deblocking_filter) {
			output_H263_EdgeFilter(output_H263_curr_enhancement_recon,&output_H263_pic);
		}

		output_H263_CalculateStatistics(output_H263_curr_image, output_H263_curr_enhancement_recon, NULL, NULL, &output_H263_bits, output_H263_QP, &output_H263_pic);

		output_H263_mv_outside_frame = output_H263_mv_outside_frame_old;
      
		/* Free the past prev_enhancement_image and prev_enhancement_recon, 
		 * no longer needed */
		if (NULL != output_H263_prev_enhancement_image) {
			output_H263_FreeImage(output_H263_prev_enhancement_image);
			output_H263_FreeImage(output_H263_prev_enhancement_recon);
		}

		/* Set new prev_enhancement_image and prev_enhancement_recon to old 
		 * next_P_image and  next_P_recon respectively */
		output_H263_prev_enhancement_image = output_H263_curr_image;
		output_H263_prev_enhancement_recon = output_H263_curr_enhancement_recon;

		if (output_H263_scalability_mode >= 3) {
			output_H263_pic.source_format = output_H263_source_format_old;
		}

		break;

	case PCT_EP:

		output_H263_EPframes++;
        
		output_H263_pels = output_H263_enhancement_pels;
		output_H263_lines = output_H263_enhancement_lines;
		output_H263_cpels = output_H263_enhancement_pels/2;
		
		/* QP fixed for enhancement layer */
		output_H263_QP = output_H263_enhancement_layer_quant;
		output_H263_pic.DQUANT = 0;

		output_H263_pic.UFEP = 1;
		output_H263_pic.RLNUM = BASE_LAYER_NUM;
		output_H263_pic.ELNUM = ENHANCEMENT_LAYER_NUM;
         
		/* Alternate picture rounding type */
		output_H263_pic.RTYPE = output_H263_rtype[output_H263_pic.ELNUM-1] % 2;
		output_H263_rtype[output_H263_pic.ELNUM-1]++;
       
		/* For all possible spatial scalings */
		if (output_H263_scalability_mode >= 3) {
			/* No MVs from reference layer, so don't need to frame recon */
			output_H263_curr_reference_recon = output_H263_UpsampleReferenceLayerPicture(output_H263_prev_recon);

			/* Enhancement layer image is temporally simultaneous, same image as was
			 * just coded in reference layer, i.e. stored as prev_image */
			output_H263_curr_image = output_H263_UpsampleReferenceLayerPicture(output_H263_prev_image);
			
			output_H263_source_format_old = output_H263_pic.source_format;
			output_H263_pic.source_format = (output_H263_scalability_mode>=7) ? output_H263_source_format_old+1 : SF_CUSTOM; 
		} else {
			output_H263_curr_reference_recon = output_H263_prev_recon;
			output_H263_curr_image = output_H263_prev_image;
		}

		output_H263_curr_enhancement_recon = output_H263_InitImage(output_H263_pels*output_H263_lines);
       
		output_H263_mv_outside_frame_old = output_H263_mv_outside_frame;
		output_H263_mv_outside_frame = ON;
        
		/* Code the current EP picture. */
		output_H263_CodeOneEP(output_H263_curr_image, output_H263_prev_enhancement_image, output_H263_prev_enhancement_recon, output_H263_QP, output_H263_P_frameskip*output_H263_orig_frameskip, &output_H263_bits, &output_H263_pic, output_H263_curr_reference_recon, output_H263_curr_enhancement_recon);

		if (output_H263_deblocking_filter) {
			output_H263_EdgeFilter(output_H263_curr_enhancement_recon,&output_H263_pic);
		}

		output_H263_CalculateStatistics(output_H263_curr_image, output_H263_curr_enhancement_recon, NULL, NULL, &output_H263_bits, output_H263_QP, &output_H263_pic);

		if (NULL != output_H263_prev_enhancement_image) {  
			output_H263_FreeImage(output_H263_prev_enhancement_image);
			output_H263_FreeImage(output_H263_prev_enhancement_recon);
		}

		output_H263_prev_enhancement_image = output_H263_curr_image;
		output_H263_prev_enhancement_recon = output_H263_curr_enhancement_recon;

		output_H263_mv_outside_frame = output_H263_mv_outside_frame_old;
        
		if (output_H263_scalability_mode >= 3) {
			output_H263_pic.source_format = output_H263_source_format_old;
		}

		break;

	default:

		break;
	}

	/* Store frame and update Thread counter and image counter for VRC */
	if (output_H263_pic.reference_picture_selection) {
		if (!output_H263_pic.sync) {
			/* for sync frames, only store the first thread */
			output_H263_StorePicture(output_H263_pic.TR, output_H263_prev_image, output_H263_prev_recon);
		}
		/* store the TRP for the next picture */
		output_H263_thread[output_H263_thread_count][(output_H263_frame_in_thread+1)%output_H263_frames_per_thread]->TRP = output_H263_pic.TR;

		++output_H263_thread_count;
		if (output_H263_thread_count == output_H263_number_of_threads || output_H263_pic.picture_coding_type == PCT_INTRA) {
			/* If this frame is a sync Intra frame or all threads or coded
			   we increment the picture in all threads */
			/* The TRP of remaining threads for I reference pictures 
			   should be updated */
			for (j=output_H263_thread_count;j<output_H263_number_of_threads;j++)
				output_H263_thread[j][(output_H263_frame_in_thread+1)%output_H263_frames_per_thread]->TRP = output_H263_pic.TR;
			output_H263_thread_count = 0;
			output_H263_frame_in_thread = ++output_H263_frame_in_thread % output_H263_frames_per_thread;
		}
	}

	/* Update full extended PTYPE  */
	if (output_H263_EPTYPE) {
		/* UFEP is set once every 5 sec or once every five pictures *
		 * whichever is a larger time interval                       */
		if ((output_H263_frames % 5 && (output_H263_TR_old - output_H263_TR_new) / output_H263_pic.src_frame_rate >= 5) || (output_H263_TR_old - output_H263_TR_new) / output_H263_pic.src_frame_rate >= 5) {
			output_H263_TR_new = output_H263_pic.TRB;
			output_H263_pic.UFEP = ON;
		} else {
			output_H263_pic.UFEP = OFF;
		}
		output_H263_TR_old = output_H263_pic.TRB; 
	}

	/* Can immediately free true B images, as the are never used for prediction */
	if (output_H263_successive_B_frames && PCT_B == output_H263_pic.picture_coding_type) {
		output_H263_FreeImage(output_H263_curr_image);
		output_H263_FreeImage(output_H263_curr_recon);
	}

	if (output_H263_pic.PB) {
		output_H263_FreeImage(output_H263_B_image);
		output_H263_FreeImage(output_H263_B_recon);
	}

	/* P_frameskip = (successive_B_frames) ? (successive_B_frames+1)*fixed_frameskip : frameskip; */
	output_H263_P_frameskip = output_H263_frameskip;
	output_H263_prev_pict_type = output_H263_pic.picture_coding_type;

	output_H263_DeterminePictureType(&output_H263_frame_no, &output_H263_pic, output_H263_P_frameskip, output_H263_True_B_frameskip);
	if (output_H263_pic.picture_coding_type != output_H263_prev_pict_type) ++output_H263_pic.GFID;

	output_H263_pic.RTYPE = 0;

	/* Set the reference picture according to the thread and frame in thread */
	/* frame in thread is reset to zero after a sync frame */
	if (output_H263_pic.reference_picture_selection) {
		output_H263_pic.TRPI = YES;
		output_H263_pic.TRP = output_H263_thread[output_H263_thread_count][output_H263_frame_in_thread]->TRP;
		output_H263_pic.RPSMF = NO_ACK_NACK;
		output_H263_pic.BCM->present = NO;
		output_H263_thread[output_H263_thread_count][output_H263_frame_in_thread]->TR = output_H263_pic.TR;
		if (output_H263_pic.picture_coding_type != PCT_INTRA) {
			if (!output_H263_frame_in_thread && output_H263_thread_count != 0) {
				/* this is a sync frame: code the same TR again
				   if GOBs are used, only the a certain number of GOBs 
				   will be encoded as sync GOBs (not the whole picture) */
				output_H263_frame_no = output_H263_prev_I_P_frame_no = output_H263_prev_frame_no;
				output_H263_pic.TR = output_H263_prev_I_P_pic_TR = output_H263_thread[0][output_H263_frame_in_thread]->TR;
				output_H263_pic.sync = 1;
			} else {
				output_H263_pic.sync = 0;
			}
			output_H263_GetReferencePicture(output_H263_pic.TRP, output_H263_prev_image, output_H263_prev_recon);
		} else {
			output_H263_thread_count = output_H263_frame_in_thread = 0;
		}
		/* previous frame number is needed to code a sync frame */
		output_H263_prev_frame_no = output_H263_frame_no;
	}
#ifdef VISI
	if (*output_H263_vectors) {
		int i, c = 0, nx = 0, ny = 0, gx = 0, gy =0, xi, yi;
		if (output_H263MV_bcur) {
			for (i = 0; i < output_H263MV_bcur; i++) {
				gx += output_H263MV_buf[i][0];
				gy += output_H263MV_buf[i][1];
			}
/*			gx /= i;
			gy /= i;
			if (gx >= output_H263_pels/2) xi = 1;
			else xi = -1;
			if (gy >= output_H263_lines/2) yi = 1;
			else yi = -1;
			for (i = 0; i < output_H263MV_bcur; i++) {
				if (output_H263MV_buf[i][0] > gx-48
					&& output_H263MV_buf[i][0] < gx+48
					&& output_H263MV_buf[i][1] > gy-32
					&& output_H263MV_buf[i][1] < gy+32) {
					nx += output_H263MV_buf[i][0];
					ny += output_H263MV_buf[i][1];
					c++;
					if (xi*(output_H263MV_buf[i][0]-gx) > yi*(gy-output_H263MV_buf[i][1])) {
						gx = output_H263MV_buf[i][0];
						gy = output_H263MV_buf[i][1];
					}
				}
			}
			nx += 2*gx;
			ny += 2*gy;
			c++; c++;
*/
			c = i;
			nx = gx;
			ny = gy;
		}

		if (c > 3) {
			if ((output_H263MV_vec[VEC_MAX-1][0]-nx/c) + (output_H263MV_vec[VEC_MAX-1][1]-ny/c) >
				output_H263_pels/(picnr-output_H263MV_last) + output_H263_lines/(picnr-output_H263MV_last)) {
				memmove(output_H263MV_vec[1],output_H263MV_vec[2],sizeof(output_H263MV_vec)-2*sizeof(output_H263MV_vec[0]));
				output_H263MV_vec[VEC_MAX-1][0] = 0;
				output_H263MV_vec[VEC_MAX-1][1] = 0;
			}
			memmove(output_H263MV_vec[1],output_H263MV_vec[2],sizeof(output_H263MV_vec)-2*sizeof(output_H263MV_vec[0]));
			output_H263MV_vec[VEC_MAX-1][0] = nx/c;
			output_H263MV_vec[VEC_MAX-1][1] = ny/c;
			output_H263MV_last = picnr;
			output_H263MV_fd = open(output_H263_vectors,O_WRONLY|O_BINARY|O_CREAT|O_TRUNC,0666);
			write(output_H263MV_fd,output_H263MV_vec,sizeof(output_H263MV_vec));
			close(output_H263MV_fd);
		}
	}
#endif
	return 1;
}

#define SAFE_output_H263_FreeImage(x) do {if (x) {output_H263_FreeImage(x); x = NULL;}} while(0)

void output_H263_stop()
{
	SAFE_output_H263_FreeImage(output_H263_prev_P_image);
	SAFE_output_H263_FreeImage(output_H263_prev_P_recon);
	SAFE_output_H263_FreeImage(output_H263_next_P_image);
	SAFE_output_H263_FreeImage(output_H263_next_P_recon);
	SAFE_output_H263_FreeImage(output_H263_prev_enhancement_image);
	SAFE_output_H263_FreeImage(output_H263_prev_enhancement_recon);
	SAFE_output_H263_FreeImage(output_H263_curr_image);
	SAFE_output_H263_FreeImage(output_H263_curr_recon);

	transport_wstop();
#ifdef VISI
	if (output_H263MV_fd >= 0) {
		close(output_H263MV_fd);
		output_H263MV_fd = -1;
	}
#endif
}

/**********************************************************************
 *
 *	Name:         DeterminePictureType
 *	Description:  Sets next picture type to be encoded 
 *	
 *	Input:        frame number, picture structure, frameskip (P and B)
 *        
 *	Returns:      
 *	Side effects: 
 *
 *	Date: 970930  Author: Michael Gallant <mikeg@ee.ubc.ca>
 *
 ***********************************************************************/

void output_H263_DeterminePictureType(int *frame_no, Pict *pic, int P_skip, int B_skip)
{
	static int i = 0;
	static int last_P_skip;

	i++;
//	printf("picnr %i\n",i);
	if (output_H263_force_intra && !(i%output_H263_force_intra)) {
//		printf("intra\n");
		*frame_no = output_H263_prev_I_P_frame_no = output_H263_prev_I_P_frame_no + P_skip;
		output_H263_prev_I_P_pic_TR = pic->TR = output_H263_prev_I_P_pic_TR + (P_skip * output_H263_orig_frameskip);
		pic->picture_coding_type = PCT_INTRA;
	} else {
//		printf("other\n");
		if ((output_H263_scalability_mode) && (PCT_INTER == output_H263_prev_pict_type || PCT_INTRA == output_H263_prev_pict_type) ) {
			/* EI or EP pictures */
			pic->picture_coding_type = (PCT_INTRA == output_H263_prev_pict_type) ? PCT_EI : PCT_EP;
			*frame_no = output_H263_prev_I_P_frame_no;
			pic->TR = output_H263_prev_I_P_pic_TR;
		} else {
			if (output_H263_successive_B_frames) {
				if (output_H263_B_pictures_to_code) {
					/* B picture */
					pic->picture_coding_type = PCT_B;
					*frame_no = output_H263_prev_I_P_frame_no - (output_H263_B_pictures_to_code * B_skip);
					--output_H263_B_pictures_to_code;
					pic->TR = output_H263_prev_I_P_pic_TR + ( (output_H263_successive_B_frames - output_H263_B_pictures_to_code) * (B_skip * output_H263_orig_frameskip));
					pic->TRB = pic->TR - output_H263_prev_I_P_pic_TR;
					output_H263_TRP = last_P_skip;
				} else {
					/* P picture */
					if (PCT_B == output_H263_prev_pict_type) { 
						output_H263_prev_I_P_pic_TR = pic->TR + B_skip;
					}
					output_H263_B_pictures_to_code = output_H263_successive_B_frames;
					if (output_H263_prev_I_P_frame_no + P_skip <= output_H263_end) {
						pic->picture_coding_type = PCT_INTER;
						*frame_no = output_H263_prev_I_P_frame_no = output_H263_prev_I_P_frame_no + P_skip;
						pic->TR = output_H263_prev_I_P_pic_TR + (P_skip * output_H263_orig_frameskip);
						last_P_skip = P_skip;
					} else {
						*frame_no = output_H263_prev_I_P_frame_no;
						pic->picture_coding_type = -1;
						pic->TR = output_H263_prev_I_P_pic_TR;
					}
				}
			} else {
				if (output_H263_pb_frames || output_H263_improved_pb_frames) {
					if (output_H263_prev_I_P_frame_no + P_skip > output_H263_end) {
						*frame_no = output_H263_prev_I_P_frame_no;
						pic->picture_coding_type = -1;
						pic->TR = output_H263_prev_I_P_pic_TR;
					} else {
						if (PCT_INTRA == output_H263_prev_pict_type || P_skip>8 || (output_H263_prev_I_P_frame_no + 2*P_skip) > output_H263_end) {
							/* If previous picture is INTRA || the temporal reference of P thus
							   the temporal reference of B is larger than 8 || there are not enough
							   frames left to code as PB than code the next picture as P picture */
							pic->PB = 0;
							pic->picture_coding_type = PCT_INTER;   
							*frame_no = output_H263_prev_I_P_frame_no = output_H263_prev_I_P_frame_no + P_skip;
							output_H263_prev_I_P_pic_TR = pic->TR = output_H263_prev_I_P_pic_TR + (P_skip * output_H263_orig_frameskip);
						} else { /* PB or IPB picture */
							pic->PB = (output_H263_improved_pb_frames) ? IM_PB_FRAMES : PB_FRAMES;
							pic->picture_coding_type = (output_H263_improved_pb_frames) ? PCT_IPB : PCT_PB;
							*frame_no = output_H263_prev_I_P_frame_no + P_skip;
							output_H263_prev_I_P_frame_no += 2 * P_skip;
							output_H263_prev_I_P_pic_TR = pic->TR = output_H263_prev_I_P_pic_TR + (2 * P_skip * output_H263_orig_frameskip);
						}
					}
				} else {
					/* P picture */
					pic->picture_coding_type = PCT_INTER;
					*frame_no = output_H263_prev_I_P_frame_no = output_H263_prev_I_P_frame_no + P_skip;
					output_H263_prev_I_P_pic_TR = pic->TR = output_H263_prev_I_P_pic_TR + (P_skip * output_H263_orig_frameskip);
				}
			}
		}
	}
}

/**********************************************************************
 *
 *	Name:         CalculateStatistics
 *	Description:  Displays bit usage and SNR info
 *	
 *	Input:        current image and reconstructed image, bit structure
 *                    and picture structure
 *        
 *	Returns:      
 *	Side effects: 
 *
 *	Date: 970930  Author: Michael Gallant <mikeg@ee.ubc.ca>
 *
 ***********************************************************************/
void output_H263_CalculateStatistics(PictImage *curr_image, PictImage *curr_recon, PictImage *pb_b_image, PictImage *pb_b_recon, Bits *bits, int QP, Pict *pic)
{
	output_H263_ComputeSNR(output_H263_curr_image, output_H263_curr_recon, &output_H263_res, pic->picture_coding_type);

	if (output_H263_arith_used) {
		bits->header += output_H263_encoder_flush();
		output_H263_arith_used = 0;
	}

	bits->header += output_H263_alignbits();  /* pictures shall be byte aligned */

	output_H263_AddBitsPicture(bits);
      
	switch (pic->picture_coding_type) {
	case PCT_INTRA:
		output_H263_AddRes(&output_H263_intra_res,&output_H263_res,pic);
		output_H263_AddBits(&output_H263_intra_bits, bits);
		break;

	case PCT_INTER:
		output_H263_AddRes(&output_H263_inter_res,&output_H263_res,pic);

		output_H263_AddBits(&output_H263_inter_bits, bits);
		output_H263_AddBits(&output_H263_total_bits, bits);  
		break;

	case PCT_PB:
	case PCT_IPB:
		output_H263_AddRes(&output_H263_inter_res,&output_H263_res,pic);
		output_H263_ComputeSNR(pb_b_image, pb_b_recon, &output_H263_res, pic->picture_coding_type);
		output_H263_AddRes(&output_H263_pb_res,&output_H263_res,pic);

		output_H263_AddBits(&output_H263_pb_bits, bits);
		output_H263_AddBits(&output_H263_total_bits, bits); 
		break;

	case PCT_B:
		output_H263_AddRes(&output_H263_b_res,&output_H263_res,pic);
		output_H263_AddRes(&output_H263_scal_res,&output_H263_res,pic);

		output_H263_AddBits(&output_H263_b_bits, bits);
		output_H263_AddBits(&output_H263_scal_bits, bits); 
		break;

	case PCT_EI:
		output_H263_AddRes(&output_H263_ei_res,&output_H263_res,pic);
		output_H263_AddRes(&output_H263_scal_res,&output_H263_res,pic);
		
		output_H263_AddBits(&output_H263_ei_bits, bits);
		output_H263_AddBits(&output_H263_scal_bits, bits);
		break;

	case PCT_EP:
		output_H263_AddRes(&output_H263_ep_res,&output_H263_res,pic);
		output_H263_AddRes(&output_H263_scal_res,&output_H263_res,pic);

		output_H263_AddBits(&output_H263_ep_bits, bits);
		output_H263_AddBits(&output_H263_scal_bits, bits);
		break;

	default:
		break;

	}
}

int output_H263_FrameLayerRC(Pict *pic)
{
	int frameskip = 1;

	if (!output_H263_targetrate) return output_H263_chosen_frameskip;
	switch (output_H263_rate_control_method) {
	case  NO: 
		frameskip = output_H263_chosen_frameskip;
		break;
		
	case TMN5_RC:
	case TMN8_RC:
		output_H263_CommBacklog = mmax(0, output_H263_CommBacklog + (int) (output_H263_bits.total - (float) pic->bit_rate / ((pic->PB) ? (pic->target_frame_rate/2) : (pic->target_frame_rate))));
		frameskip = 1;
		while (output_H263_CommBacklog > ((float) pic->bit_rate / pic->target_frame_rate)) {
			output_H263_CommBacklog = mmax(0, (int) (output_H263_CommBacklog - (float) pic->bit_rate / pic->target_frame_rate));
			frameskip += 1;
		}
		frameskip = (int) (frameskip * pic->src_frame_rate / pic->target_frame_rate); // @@@@@@@@
		break;
		
	case OFFLINE_RC:
		/* Aim for the targetrate with a once per frame rate control scheme */
		/* when generating the MPEG-4 anchors, rate control was started after 70% of the sequence was finished. Set start_rate_control with option "-R <n>" */
		if (output_H263_frame_no - output_H263_startf > (output_H263_end - output_H263_startf) * output_H263_start_rate_control/100.0) 
			pic->QUANT = output_H263_FrameUpdateQP(output_H263_total_bits.total + output_H263_intra_bits.total, output_H263_bits.total, (output_H263_end-output_H263_frame_no) / output_H263_chosen_frameskip + ((pic->PB) ? 1:0), output_H263_QP, output_H263_targetrate, output_H263_seconds);
		frameskip = output_H263_chosen_frameskip;
		break;      
	default:
		break;
	}
	
	return frameskip;
}


#ifdef VISI
void output_H263MV_put(int x, int y, int xdest, int ydest)
{
	if ((x!=xdest || y!=ydest) && output_H263MV_bcur >= 0) {
//		output_H263MV_buf[output_H263MV_bcur][0] = x;
//		output_H263MV_buf[output_H263MV_bcur][1] = y;
		output_H263MV_buf[output_H263MV_bcur][0] = xdest+8;
		output_H263MV_buf[output_H263MV_bcur][1] = ydest+8;
		output_H263MV_bcur++;
		if (output_H263MV_bcur >= MVBUF_MAX) output_H263MV_bcur = -1;
	}
}
#endif
