/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "../config.h"

#ifdef HAVE_LIBX11
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <unistd.h>
#include <limits.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#ifdef HAVE_LIBXEXT
#include <X11/extensions/XShm.h>
#endif

#define TRANSPORT_INTERNAL
#include "version.h"
#include "error.h"
#include "configuration.h"
#include "color.h"
#include "transport.h"
#define SYMBOL(x) output_X_ ## x

/************ Prototypes ************/

int output_X_error(Display * d, XErrorEvent * xeev);
int output_X_ioerror(Display *d);
void output_X_handle();
void output_X_start(const char *unused);
void output_X_stop(void);
int output_X_write(unsigned char *buffer, int picnr);
void output_X_abort(int x);

/************************************/

int output_X_fmts() { return 1; }
const char *output_X_fmt(int nr) { return "X"; }

#define OUTPUT_X_EVENTMASK (KeyPressMask)

char *output_X_buffer = NULL;
#ifdef HAVE_LIBXEXT
XShmSegmentInfo output_X_shminfo;
#endif
Display *output_X_disp = NULL;
Window output_X_win = 0;
XImage *output_X_ximage = NULL;
GC output_X_gc = 0;
int output_X_errorcnt = 0;
int output_X_noerr = 0;
Colormap output_X_colormap = 0;
static Atom output_X_WM_PROTOCOLS = 0, output_X_WM_DELETE_WINDOW = 0;
const char *output_X_color_fmt = "";
int output_X_bytes = 0, output_X_width = 0, output_X_height = 0;
char output_X_display[4096] = "";

START_DATA
INTOPTION(width,0,INT_MAX);
INTOPTION(height,0,INT_MAX);
OPTION(color_fmt) NOSET O_GET {
	int screen, dpy_depth;
	Display *display;
	Visual *visual;
	ASSERT((display = XOpenDisplay(*output_X_display?output_X_display:NULL)),(MSG("open display failed\n")));

	screen = DefaultScreen(display);
	visual = DefaultVisual(display, screen);
	dpy_depth = DefaultDepth(display, screen);

	if (visual->class == TrueColor) {
		if (dpy_depth == 24) output_X_color_fmt = "RGB24";
		else if (dpy_depth == 16) output_X_color_fmt = "RGB16";
		else {
			ERROR(MSG("%i bpp for TrueColor not supported"),dpy_depth);
			END(1);
		}
	} else if (visual->class == PseudoColor) {
		ASSERT(dpy_depth == 8,(MSG("only 8 bpp for PseudoColor supported")));
		output_X_color_fmt = "PAL8";
	}
	XCloseDisplay(display);
	strcpy(value,output_X_color_fmt);
}
STROPTION(display);
END_DATA


int output_X_error(Display * d, XErrorEvent * xeev)
{
	output_X_errorcnt = 1;
	return 0;
}

int output_X_ioerror(Display *d)
{
	if (!output_X_noerr) {
		ERROR(MSG("fatal X Windows error"));
		END(1);
	}
	return 0;
}

void output_X_handle()
{
	XEvent xev;
	KeySym keysym;

	while (XCheckWindowEvent(output_X_disp, output_X_win, OUTPUT_X_EVENTMASK, &xev)) {
		switch (xev.type) {
		case KeyPress:
			keysym = XKeycodeToKeysym(output_X_disp, xev.xkey.keycode, 0);
			if (keysym == XK_Q || keysym == XK_q) {
				ERROR(NOERR("user abort"));
				END(0);
			}
			break;

		case ClientMessage:
			if (xev.xclient.message_type == output_X_WM_PROTOCOLS && xev.xclient.data.l[0] == output_X_WM_DELETE_WINDOW) {
				ERROR(NOERR("user abort"));
				END(0);
			}
			break;

		default:
			break;

		}
	} 
}


void output_X_start(const char *unused)
{
	static XGCValues values;
	XSizeHints hints;
	XWMHints wmhints;
	XTextProperty windowName, iconName;
	Window root;
	int screen_num;
	int depth;
	static XColor palette[256];

	int i = 0, r, g, b;
	int tg[9] = {0,32,64,96,128,160,192,224,255};
	int trb[5] = {0,64,128,192,255};
	char *name = "bttvgrab " VERSION;

	output_X_bytes = color_bytes(color_val(output_X_color_fmt),output_X_width,output_X_height);
#ifdef HAVE_LIBXEXT
	output_X_shminfo.shmaddr = 0;
	output_X_shminfo.shmid = -1;
#endif

	ASSERT(signal(SIGPIPE,output_X_abort) != SIG_ERR,(ERRMSG("signal")));

	ASSERT((output_X_disp = XOpenDisplay(NULL)),(MSG("open display failed\n")));

	output_X_noerr = 0;
	output_X_errorcnt = 0;
	XSetErrorHandler(output_X_error);
	XSetIOErrorHandler(output_X_ioerror);
	
	screen_num = DefaultScreen(output_X_disp);
	
	root = DefaultRootWindow(output_X_disp);
	
	output_X_win = XCreateSimpleWindow(output_X_disp, root, 0, 0, output_X_width, output_X_height, 0,
					   XWhitePixel(output_X_disp,screen_num), XBlackPixel(output_X_disp,screen_num));
	ASSERT(output_X_win,(MSG("create window failed\n")));
	
	hints.flags = PSize|PMaxSize|PMinSize;
	hints.min_width = hints.max_width = hints.width = output_X_width;
	hints.min_height = hints.max_height = hints.height = output_X_height;
	wmhints.input = True;
	wmhints.flags = InputHint;
	
	XStringListToTextProperty(&name, 1 ,&windowName);
	XStringListToTextProperty(&name, 1 ,&iconName);

	XSetWMProperties(output_X_disp, output_X_win, &windowName, &iconName, NULL, 0, &hints, &wmhints, NULL);
	
	output_X_WM_DELETE_WINDOW = XInternAtom(output_X_disp, "WM_DELETE_WINDOW", False);
	output_X_WM_PROTOCOLS = XInternAtom(output_X_disp, "WM_PROTOCOLS", False);
	XSetWMProtocols(output_X_disp, output_X_win, &output_X_WM_DELETE_WINDOW, 1);

	XSelectInput(output_X_disp, output_X_win, OUTPUT_X_EVENTMASK);

	XMapRaised(output_X_disp, output_X_win);
	
	output_X_gc = XCreateGC(output_X_disp, output_X_win, 0, &values);
	
	XSync(output_X_disp, 0);
	
	depth = DefaultDepthOfScreen(DefaultScreenOfDisplay(output_X_disp));

#ifdef HAVE_LIBXEXT	
	if (XShmQueryExtension(output_X_disp)) {
		output_X_ximage = XShmCreateImage(output_X_disp, DefaultVisual(output_X_disp, screen_num), 
						  depth, ZPixmap, NULL, &output_X_shminfo, output_X_width, output_X_height);
		if (output_X_ximage) {
			output_X_shminfo.shmid = shmget(IPC_PRIVATE, output_X_ximage->bytes_per_line*output_X_ximage->height, IPC_CREAT|0777);
			if (output_X_shminfo.shmid < 0) {
				XDestroyImage(output_X_ximage);
				output_X_ximage = NULL;
			}
		}
		if (output_X_ximage) {
			output_X_buffer = output_X_ximage->data = output_X_shminfo.shmaddr = shmat(output_X_shminfo.shmid, 0, 0);
			if (output_X_buffer == (char *)-1) {
				output_X_shminfo.shmaddr = output_X_buffer = NULL;
				XDestroyImage(output_X_ximage);
				output_X_ximage = NULL;
			}
		}
		if (output_X_ximage) {
			XShmAttach(output_X_disp, &output_X_shminfo);
			XSync(output_X_disp, 0);
			if (output_X_errorcnt) {
				shmdt(output_X_shminfo.shmaddr);
				output_X_shminfo.shmaddr = output_X_buffer = NULL;
				XDestroyImage(output_X_ximage);
				output_X_ximage = NULL;
			}
		}
	}
#endif
	if (!output_X_ximage) {
		SAFE_MALLOC(output_X_buffer,((depth + 7) / 8) * output_X_width * output_X_height);
		output_X_ximage = XCreateImage(output_X_disp, DefaultVisual(output_X_disp, screen_num), 
					       depth, ZPixmap, 0, output_X_buffer, output_X_width, output_X_height, depth,
					       output_X_width * ((depth + 7) / 8));
		ASSERT(output_X_ximage,(MSG("CreateImage Failed (%i bpp)\n"),depth));
	}

	if(depth == 8) {
		output_X_colormap = XCreateColormap(output_X_disp, DefaultRootWindow(output_X_disp), DefaultVisual(output_X_disp, screen_num), AllocAll);
		i = 16;
		for (g=0; g<9; g++) for (r=0; r<5; r++) for (b=0;b<5;b++) {
			palette[i].pixel=i;
			palette[i].red = trb[r]*256;
			palette[i].green = tg[g]*256;
			palette[i].blue = trb[b]*256;
			palette[i].flags=DoRed|DoGreen|DoBlue;
			i++;
		}
		XStoreColors(output_X_disp, output_X_colormap, palette, 256);
		XSetWindowColormap(output_X_disp, output_X_win, output_X_colormap);
	}
	XFlush(output_X_disp);
}

void output_X_stop(void)
{
	output_X_noerr = 1;
	if (output_X_ximage) {
		XDestroyImage(output_X_ximage);
		output_X_ximage = NULL;
	}
#ifdef HAVE_LIBXEXT
	if (output_X_shminfo.shmid >= 0) {
		if (output_X_shminfo.shmaddr) {
			XShmDetach(output_X_disp, &output_X_shminfo);
			shmdt(output_X_shminfo.shmaddr);
			output_X_shminfo.shmaddr = output_X_buffer = NULL;
		}
		shmctl(output_X_shminfo.shmid, IPC_RMID, 0);
		output_X_shminfo.shmid = -1;
	}
#endif
	SAFE_FREE(output_X_buffer);

	if (output_X_win) {
		XEvent xev;
		if (output_X_gc) {
			XFreeGC(output_X_disp, output_X_gc);
			output_X_gc = 0;	
		}
		XUnmapWindow(output_X_disp,output_X_win);
		XFlush(output_X_disp);
		while (XCheckWindowEvent(output_X_disp, output_X_win, OUTPUT_X_EVENTMASK, &xev));
		XDestroyWindow(output_X_disp,output_X_win);
		output_X_win = 0;
	}
	if (output_X_colormap) {
		XFreeColormap(output_X_disp,output_X_colormap);
		output_X_colormap = 0;
	}
	if (output_X_disp) {
		XCloseDisplay(output_X_disp);
		output_X_disp = NULL;
	}
}

int output_X_write(unsigned char *buffer, int picnr)
{
	char msg[80];
	memcpy(output_X_buffer,buffer,output_X_bytes);
	sprintf(msg,"displaying frame [%i]",picnr);
	(*transport_otrace)(msg);

#ifdef HAVE_LIBXEXT
	if (output_X_shminfo.shmaddr) XShmPutImage(output_X_disp, output_X_win, output_X_gc, output_X_ximage, 0, 0, 0, 0, output_X_width, output_X_height, False);
	else
#endif
		XPutImage(output_X_disp, output_X_win, output_X_gc, output_X_ximage,  0, 0, 0, 0, output_X_width, output_X_height);

	XFlush(output_X_disp);
	output_X_handle();
	return 1;
}

void output_X_abort(int x)
{
	ERROR(MSG("connection to server lost"));
	END(0);
}
#endif
