/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/
#include "../config.h"

#ifdef HAVE_LIBGIF
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <gif_lib.h>
#include "transport.h"
#include "error.h"
#include "configuration.h"
#define SYMBOL(x) output_gif_ ## x

ColorMapObject *output_gif_map = NULL;
int output_gif_width = 0;
int output_gif_height = 0;

START_DATA
INTOPTION(width,0,INT_MAX);
INTOPTION(height,0,INT_MAX);
CONSTINFO(color_fmt,"PAL8");
END_DATA

/************ Prototypes ************/

void output_gif_start(const char *unused);
int output_gif_write(unsigned char *buffer, int picnr);
void output_gif_stop();

/************************************/

int output_gif_fmts() { return 1; }
const char *output_gif_fmt(int nr) { return "gif"; }

void output_gif_start(const char *unused)
{
	int tg[9] = {0,32,64,96,128,160,192,224,255};
	int trb[5] = {0,64,128,192,255};
	int i = 16, r, g, b;

	if (!output_gif_map) {
		NOTNULL(output_gif_map = MakeMapObject(256,NULL),(ERRMSG("gif init")));
		for (g=0; g<9; g++) for (r=0; r<5; r++) for (b=0;b<5;b++) {
			output_gif_map->Colors[i].Red = trb[r];
			output_gif_map->Colors[i].Green = tg[g];
			output_gif_map->Colors[i].Blue = trb[b];
			i++;
		}
	}

	transport_wstart("gif",output_gif_width,output_gif_height,1);
}

int output_gif_write(unsigned char *buffer, int picnr)
{
	GifFileType *gif;
	int fh;

	fh = transport_wframe(picnr);
	NOTNULL(gif = EGifOpenFileHandle(fh),(ERRMSG("EGifOpenFileHandle")));
	ASSERT(EGifPutScreenDesc(gif,output_gif_width,output_gif_height,256,0,output_gif_map) == GIF_OK,
	       (ERRMSG("EGifPutScreenDesc")));
	ASSERT(EGifPutImageDesc(gif,0,0,output_gif_width,output_gif_height,0,output_gif_map) == GIF_OK,
	       (ERRMSG("EGifPutImageDesc")));
	ASSERT(EGifPutLine(gif,buffer,output_gif_width*output_gif_height) == GIF_OK,
	       (ERRMSG("EGifPutLine")));
	ASSERT(EGifCloseFile(gif) == GIF_OK,(ERRMSG("EGifCloseFile")));
	return 1;
}

void output_gif_stop()
{
	if (output_gif_map) {
		FreeMapObject(output_gif_map);
		output_gif_map = NULL;
	}
}
#else
#warning gif output not available
#endif
