/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/
#include "../config.h"

#ifdef HAVE_LIBAA
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <limits.h>
#include <aalib.h>
#include "error.h"
#include "configuration.h"
#define SYMBOL(x) output_term_ ## x

/************ Prototypes ************/

void output_term_start(const char *ignored);
int output_term_write(unsigned char *buffer, int picnr);
void output_term_stop();

/************************************/

int output_term_fmts() { return 1; }
const char *output_term_fmt(int nr) { return "term"; }

aa_context *output_term_context = NULL;
unsigned char *output_term_buffer = NULL;
aa_renderparams output_term_params;
int output_term_width = 0, output_term_height = 0;
int output_term_bright = 8, output_term_contrast = 48, output_term_gamma = 190;

START_DATA
INTOPTION(width,0,INT_MAX);
INTOPTION(height,0,INT_MAX);
CONSTINFO(color_fmt,"GREY8");
INTOPTION(bright,0,100);
INTOPTION(contrast,0,1000);
INTOPTION(gamma,0,100);
END_DATA

void output_term_start(const char *ignored)
{
	ASSERT(!(output_term_width%2 || output_term_height%2),
	       (MSG("width and height must be multiples of 2 for term output")));
	aa_defparams.width = output_term_width / 2;
	aa_defparams.height = output_term_height / 2;
	NOTNULL((output_term_context = aa_autoinit(&aa_defparams)),(MSG("AA-Lib initialization")));
	output_term_buffer = aa_image(output_term_context);
	output_term_params.bright = output_term_bright;
	output_term_params.contrast = output_term_contrast;
	output_term_params.gamma = ((float)output_term_gamma)/100.0;
	output_term_params.dither = 2;
	output_term_params.inversion = 0;
	output_term_params.randomval = 0;

}

int output_term_write(unsigned char *buffer, int picnr)
{
	memcpy(output_term_buffer,buffer,output_term_width*output_term_height);
	aa_render(output_term_context, &output_term_params, 0, 0, aa_scrwidth(output_term_context), aa_scrheight(output_term_context));
	aa_flush(output_term_context);
	return 1;
}

void output_term_stop()
{
	if (output_term_context) {
		aa_close(output_term_context);
		output_term_context = NULL;
	}
}
#endif
