/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "../config.h"

#ifndef _WIN32
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <pthread.h>
#ifdef HAVE_SEMAPHORE_H
#include <semaphore.h>
#endif
#include <errno.h>
#include "error.h"
#include "configuration.h"
#include "read_sound.h"
#define SYMBOL(x) read_sound_dsp_ ## x
#ifdef __CYGWIN__
#define PTH_T(x) &x
#else
#define PTH_T(x) x
#endif

/************ Prototypes ************/

void read_sound_dsp_start();
void *read_sound_dsp_child_start(void *ign);
int read_sound_dsp_get(int *fbnr, int *picnr);
void read_sound_dsp_setnext(int fbnr, int picnr);
void read_sound_dsp_stop();

/************************************/

#define BUFSIZE (48000*2*4*2)/25 // (maxSamprate * maxBytespersample * maxChannels * staySafe) / fps
//#define READ_LEN ((read_sound_dsp_channels*(read_sound_dsp_bits/8)*(read_sound_dsp_rate/25))/10)
#define READ_LEN read_sound_dsp_bufsize

char read_sound_dsp_device[PATH_MAX] = "/dev/dsp";
int read_sound_dsp_rate     = 44100;
int read_sound_dsp_bits     = 16;
int read_sound_dsp_channels = 2;

int read_sound_dsp_align;
int read_sound_dsp_fdsp = 0;
pthread_t read_sound_dsp_thread = 0;
int *read_sound_dsp_picnr = NULL;
int read_sound_dsp_curpic = -1;
int read_sound_dsp_curbuf = -1;
int read_sound_dsp_bufsize = 0;

pthread_mutex_t read_sound_dsp_stmutex;
sem_t read_sound_dsp_newbuf;

START_DATA
STROPTION(device);
INTOPTION(rate,4000,48000);
INTOPTION(bits,8,16);
INTOPTION(channels,1,4);
INTINFO(bufsize);
END_DATA

void read_sound_dsp_start()
{
	int i;
	pthread_attr_t pattr;

	if (read_sound_dsp_fdsp) return;

	EQNULL((void *)sem_init(&read_sound_dsp_newbuf,0,0),(MSG("sem_init")));
	EQNULL((void *)pthread_mutex_init(&read_sound_dsp_stmutex,NULL),(MSG("pthread_mutex_init")));

	SAFE_OPEN(read_sound_dsp_fdsp,read_sound_dsp_device,O_RDONLY);

	read_sound_dsp_align = read_sound_dsp_channels * ((read_sound_dsp_bits + 7) / 8);
	read_sound_dsp_bufsize = read_sound_dsp_bits/8 * read_sound_dsp_rate/25 * read_sound_dsp_channels;
	
	SAFE_MALLOC(read_sound_dsp_picnr,(read_sound_fbs+1)*sizeof(int));
	for (i = 0; i < read_sound_fbs; i++) read_sound_dsp_picnr[i] = INT_MAX;
	read_sound_dsp_picnr[i] = INT_MIN;
	read_sound_dsp_curpic = -1;

	pthread_attr_init(&pattr);
	pthread_attr_setdetachstate(&pattr,PTHREAD_CREATE_DETACHED);
	EQNULL((void *)pthread_create(&read_sound_dsp_thread,&pattr,read_sound_dsp_child_start,NULL),(ERRMSG("pthread_create")));
	pthread_attr_destroy(&pattr);
}

void *read_sound_dsp_child_start(void *ign)
{
	int tmp, i, picnr = 0;

	tmp = 0x00ff000b; // 256 byte fragments, number of them not limited
	NOTNEG1(ioctl(read_sound_dsp_fdsp, SNDCTL_DSP_SETFRAGMENT, &tmp),(ERRMSG("fragment %x"),tmp));
	tmp = AFMT_S16_LE;
	if (read_sound_dsp_bits == 8) tmp = AFMT_U8;
	NOTNEG1(ioctl(read_sound_dsp_fdsp, SNDCTL_DSP_SETFMT, &tmp),(ERRMSG("sound format %i"),tmp));
	NOTNEG1(ioctl(read_sound_dsp_fdsp, SNDCTL_DSP_CHANNELS, &read_sound_dsp_channels),(ERRMSG("%i sound channels"),read_sound_dsp_channels));
	NOTNEG1(ioctl(read_sound_dsp_fdsp, SNDCTL_DSP_SPEED, &read_sound_dsp_rate),(ERRMSG("sampling rate %i"),read_sound_dsp_rate));
	read_sound_dsp_curbuf = 0;

loop:
	SAFE_READ(read_sound_dsp_fdsp,read_sound_buf[read_sound_dsp_curbuf],READ_LEN,"sound data");

	pthread_mutex_lock(&read_sound_dsp_stmutex);
	sem_post(&read_sound_dsp_newbuf);
	for (i = 0; i < read_sound_fbs; i++) if (read_sound_dsp_picnr[i] == INT_MAX) break;
	if (i == read_sound_fbs) {
		tmp = 0;
		for (i = 0; i < read_sound_fbs; i++) {
			if (read_sound_dsp_picnr[tmp] > read_sound_dsp_picnr[i]
			    && read_sound_dsp_picnr[i] != read_sound_dsp_curpic) tmp = i;
		}
		read_sound_dsp_curbuf = tmp;
	} else {
		read_sound_dsp_curbuf = i;
	}
	read_sound_dsp_picnr[read_sound_dsp_curbuf] = picnr++;
	pthread_mutex_unlock(&read_sound_dsp_stmutex);
	goto loop;
}

int read_sound_dsp_get(int *fbnr, int *picnr)
{
	int i, x = 0;
	*fbnr = read_sound_fbs;
	pthread_mutex_lock(&read_sound_dsp_stmutex);
	for (i = 0; i < read_sound_fbs; i++) {
		if (read_sound_dsp_picnr[i] <= read_sound_dsp_curpic) read_sound_dsp_picnr[i] = INT_MAX;
		if (read_sound_dsp_picnr[i] == *picnr) *fbnr = i; // && read_sound_dsp_picnr[i] > read_sound_dsp_picnr[*fbnr]) *fbnr = i;
		else if (read_sound_dsp_picnr[i] != INT_MAX) x = 1;
	}
	if (*fbnr == read_sound_fbs) {
		if (x) {
			pthread_mutex_unlock(&read_sound_dsp_stmutex);
			return 0;
		} else {
			pthread_mutex_unlock(&read_sound_dsp_stmutex);
			sem_wait(&read_sound_dsp_newbuf);
			return read_sound_dsp_get(fbnr, picnr);
		}
	}
	read_sound_dsp_curpic = read_sound_dsp_picnr[*fbnr];
	pthread_mutex_unlock(&read_sound_dsp_stmutex);

	while (read_sound_dsp_curbuf == *fbnr) sem_wait(&read_sound_dsp_newbuf);
	return READ_LEN;
}

void read_sound_dsp_setnext(int fbnr, int picnr)
{
//	sem_post(&read_sound_dsp_newbuf);
}

void read_sound_dsp_stop()
{
	if (read_sound_dsp_thread) {
		pthread_t x = pthread_self();
		if (!pthread_equal(PTH_T(x),PTH_T(read_sound_dsp_thread))) pthread_kill(PTH_T(read_sound_dsp_thread),SIGABRT);
		read_sound_dsp_thread = 0;
	}
	if (read_sound_dsp_fdsp > 0) {
		close(read_sound_dsp_fdsp);
		read_sound_dsp_fdsp = -1;
		EQNULL((void *)sem_destroy(&read_sound_dsp_newbuf),(MSG("sem_destroy")));
		EQNULL((void *)pthread_mutex_destroy(&read_sound_dsp_stmutex),(MSG("pthread_mutex_destroy")));
	}
	SAFE_FREE(read_sound_dsp_picnr);
}
#else
#include <string.h>
#include "error.h"


void read_sound_dsp_configure(const char *n, const char *v)
{
}

void read_sound_dsp_inquire(const char *n, char *v)
{
}

void read_sound_dsp_start()
{
	ASSERT(0,(MSG("sound not supported in this version")));
}

int read_sound_dsp_get(int *fbnr, int *picnr)
{
	ASSERT(0,(MSG("sound not supported in this version")));
	return 0;
}

void read_sound_dsp_setnext(int fbnr, int picnr)
{
}

void read_sound_dsp_stop()
{
}
#endif
