/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

// see below for adding new output modules
#include "../config.h"

#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "transport.h"
#include "error.h"
#include "configuration.h"
#define SYMBOL(x) sound_ ## x

unsigned char **sound_buf = NULL;
int sound_fbs = 0;
int sound_file_fmt = 0;
/************ Prototypes ************/

void sound_init();
void sound_start();
void sound_stop();
void sound_end();
void sound_write(int fbnr, int picnr, int len);

/************************************/
#define PROTO(symbol) \
	extern void sound_ ## symbol ## _write(unsigned char *buffer, int picnr, int len); \
	extern void sound_ ## symbol ## _start(const char *format); \
	extern void sound_ ## symbol ## _stop(); \
	extern void sound_ ## symbol ## _configure(const char *format, const char *assignment); \
	extern void sound_ ## symbol ## _inquire(const char *format, char *value);

#define ADD_MODULE(symbol) { \
	sound_ ## symbol ## _write, \
	sound_ ## symbol ## _start, \
	sound_ ## symbol ## _stop, \
	sound_ ## symbol ## _configure, \
	sound_ ## symbol ## _inquire, \
	#symbol }

struct SOUND_MODULE_STRUCT {
	void (*write)(unsigned char *buffer, int picnr, int len);
	void (*start)(const char *format);
	void (*stop)();
	void (*configure)(const char *format, const char *assignment);
	void (*inquire)(const char *format, char *value);
	const char *name;
};

int SOUND_FORMATS = 0;

////////////////////////////////////////////////////////////////////////////////
///////////////////////// output module registration ///////////////////////////
////////////////////////////////////////////////////////////////////////////////

// Step 1: add prototypes
PROTO(dsp)
PROTO(wav)
PROTO(g728)

// Step 2: fill in structure with ADD_FORMAT (supply symbol prefix)
const struct SOUND_MODULE_STRUCT sound_formats[] = {
	ADD_MODULE(wav),
	ADD_MODULE(dsp),
	ADD_MODULE(g728),
	{ NULL, NULL, NULL, NULL, NULL, NULL }
};

// Step 3: add the new object file to Makefile.in
// Step 4: create the source file, see sound_wav.c for an example

// finished
////////////////////////////////////////////////////////////////////////////////

START_OPTIONS
INTOPTION(file_fmt,0,SOUND_FORMATS-1);
OPTION(file_fmt nr) NOSET O_GET sprintf(value,"%i",SOUND_FORMATS);
OPTION(file_fmt str) O_SET {
	int i;
	for (i = 0; i < SOUND_FORMATS; i++) {
		if (!strcmp(sound_formats[i].name,value)) {
			sound_file_fmt = i;
			break;
		}
	}
	ASSERT(i < SOUND_FORMATS,(MSG("unknown sound output format '%s'"),value));
} O_GET {
	strcpy(value,sound_formats[sound_file_fmt].name);
}
INTOPTION(fbs,4,INT_MAX) if (set) { SAFE_FREE(sound_buf); SAFE_CALLOC(sound_buf,sound_fbs,sizeof(char *)); }
PTROPTION(buf,sound_fbs);
OPTION(data) O_SET (*sound_formats[sound_file_fmt].configure)(sound_formats[sound_file_fmt].name,value);
             O_GET (*sound_formats[sound_file_fmt].inquire)(sound_formats[sound_file_fmt].name,value);
END_OPTIONS

 void (*sound_write_func)(unsigned char *buffer, int picnr, int len);
 void (*sound_setnext_func)(int fbnr);
 void (*sound_stop_func)() = NULL;

void sound_init()
{
	int modulecount = 0;
//	int fmtcount;
	SOUND_FORMATS = 0;

	while (sound_formats[modulecount].start) {
//		fmtcount = (*sound_modules[modulecount].fmts)();
//		SOUND_FORMATS += fmtcount;
//		NOTNULL(sound_formats = realloc(sound_formats, SOUND_FORMATS*sizeof(struct SOUND_FORMAT_STRUCT)),
//			(MSG("realloc failed for %i bytes"),SOUND_FORMATS*sizeof(struct SOUND_FORMAT_STRUCT)));
//		while (fmtcount--) {
//			sound_formats[SOUND_FORMATS-fmtcount-1].modulenum = modulecount;
//			strcpy(sound_formats[SOUND_FORMATS-fmtcount-1].name,(*sound_modules[modulecount].fmt)(fmtcount));
//		}
		modulecount++;
	}

	SOUND_FORMATS = modulecount;
	transport_init();
}

void sound_start()
{
	ASSERT(sound_fbs,(INTMSG("sound buffers not set")));

	(*sound_formats[sound_file_fmt].start)(sound_formats[sound_file_fmt].name);

	sound_write_func = sound_formats[sound_file_fmt].write;
	sound_stop_func = sound_formats[sound_file_fmt].stop;

}

void sound_stop()
{
	if (sound_stop_func) (*sound_stop_func)();
	sound_stop_func = NULL;
}

void sound_end()
{
	sound_stop();
//	SAFE_FREE(sound_formats);
}

void sound_write(int fbnr, int picnr, int len)
{
	(*sound_write_func)(sound_buf[fbnr],picnr,len);
}
