/************************************************************************
 *
 *  coder.c, main coding engine of tmn (TMN encoder)
 *
 *  Copyright (C) 1997  University of BC, Canada
 *
 *  Contacts:
 *  Michael Gallant                   <mikeg@ee.ubc.ca>
 *  Guy Cote                          <guyc@ee.ubc.ca>
 *  Berna Erol                        <bernae@ee.ubc.ca>
 *
 *  UBC Image Processing Laboratory   http://www.ee.ubc.ca/image
 *  2356 Main Mall                    tel.: +1 604 822 4051
 *  Vancouver BC Canada V6T1Z4        fax.: +1 604 822 5949
 *
 *  Copyright (C) 1995, 1996  Telenor R&D, Norway
 *
 *  Contacts:
 *  Robert Danielsen                  <Robert.Danielsen@nta.no>
 *
 *  Telenor Research and Development  http://www.nta.no/brukere/DVC/
 *  P.O.Box 83                        tel.:   +47 63 84 84 00
 *  N-2007 Kjeller, Norway            fax.:   +47 63 81 00 76
 *
 ************************************************************************/

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any
 * license fee or royalty on an "as is" basis. The University of
 * British Columbia disclaims any and all warranties, whether
 * express, implied, or statuary, including any implied warranties
 * or merchantability or of fitness for a particular purpose.  In no
 * event shall the copyright-holder be liable for any incidental,
 * punitive, or consequential damages of any kind whatsoever arising
 * from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs
 * and user's customers, employees, agents, transferees, successors,
 * and assigns.
 *
 * The University of British Columbia does not represent or warran
 * that the programs furnished hereunder are free of infringement of
 * any third-party patents.
 *
 * Commercial implementations of H.263, including shareware, are
 * subject to royalty fees to patent holders.  Many of these patents
 * are general enough such that they are unavoidable regardless of
 * implementation design.
 *
*/


/*****************************************************************
 *
 * Modified by Pat Mulroy, BT Labs to run syntax based arithmetic
 * coding.  SAC option, H.263 (Annex E).
 *
 *****************************************************************/


#include "sim.h"
#include "../error.h"

/**********************************************************************
 *
 *	Name:        CodeOneIntra
 *	Description:	codes one image intra
 *	
 *	Input:        pointer to image, QP
 *
 *	Returns:	pointer to reconstructed image
 *	Side effects:	memory is allocated to recon image
 *
 *	Date: 940110	Author:	Karl.Lillevold@nta.no
 *
 ***********************************************************************/


PictImage *output_H263_CodeOneIntra(PictImage *curr, int QP, Bits *bits, Pict *pic)
{
	PictImage *recon;
	MB_Structure *data = (MB_Structure *)malloc(sizeof(MB_Structure));
	int *qcoeff, *pcoeff=NULL, *rcoeff, *coeff;
	int Mode = MODE_INTRA;
	int CBP,COD;
	int i,j,k;
	int *store_coeff=NULL, *store_rcoeff=NULL; /* used to do prediction in advanced intra coding */
	int newgob = 0;

	pic ->Intra_Mode = 0;
	recon = output_H263_InitImage(output_H263_pels*output_H263_lines);
	output_H263_ZeroBits(bits);
	pic->QUANT = QP;

	if (output_H263_EPTYPE) pic->UFEP = ON;   /* Update full extended PTYPE is set for Intra pictures */
	if (output_H263_advanced_intra_coding) {
		pic->UFEP = ON;   /* Update full extended PTYPE is set for Intra pictures */
		/* store the qcoeff for the frame */
		SAFE_MALLOC(store_rcoeff,sizeof(int)*384*(output_H263_pels/MB_SIZE)*(output_H263_lines/MB_SIZE));
		SAFE_MALLOC(store_coeff,sizeof(int)*384*(output_H263_pels/MB_SIZE)*(output_H263_lines/MB_SIZE));
		SAFE_MALLOC(pcoeff,sizeof(int)*384);
	}
	SAFE_MALLOC(qcoeff,sizeof(int)*384);
	SAFE_MALLOC(rcoeff ,sizeof(int)*384);

	bits->header += output_H263_CountBitsPicture(pic);

	COD = 0; /* Every block is coded in Intra frame */
	for ( j = 0; j < output_H263_lines/MB_SIZE; j++) {

		/* insert sync in *every* GOB if use_gobsync is chosen */
		if (pic->use_gobsync && j != 0) {
			bits->header += output_H263_CountBitsGOB(j,QP,pic);
			newgob = 1;
		}
		for ( i = 0; i < output_H263_pels/MB_SIZE; i++) {

			pic->MB = i + j * (output_H263_pels/MB_SIZE);
			bits->no_intra++;
			output_H263_FillLumBlock(i*MB_SIZE, j*MB_SIZE, curr, data);
			output_H263_FillChromBlock(i*MB_SIZE, j*MB_SIZE, curr, data);

			coeff = output_H263_MB_Encode(data);
			/* store the QP for every macroblock */
			output_H263_quant_map[j+1][i+1] = QP;

			if (output_H263_advanced_intra_coding) {
				/* store the DCT coefficients for encoding */
				memcpy( (void *) (store_coeff + (i + j*output_H263_pels/MB_SIZE)*384), (void *) coeff, sizeof(int) * 384);
				/* Do Intra mode prediction */
				pic->Intra_Mode = output_H263_Choose_Intra_Mode(coeff, store_coeff, i, j, newgob);
				for (k=0;k<6;k++) {

					output_H263_Intra_AC_DC_Encode(coeff, store_rcoeff, pic->Intra_Mode, i, j, newgob,k);
					output_H263_Quant_blk(coeff,pcoeff,QP,Mode,k);
					output_H263_Dequant(pcoeff, rcoeff, QP, Mode,k);
					output_H263_Intra_AC_DC_Decode(rcoeff, store_rcoeff, pic->Intra_Mode, i, j, newgob,k);
				}
				output_H263_MB_Decode(rcoeff, data);
				output_H263_Scan(pcoeff, pic->Intra_Mode);
				CBP = output_H263_FindCBP(pcoeff,Mode,64);
				if (!output_H263_syntax_arith_coding) {
					output_H263_CountBitsMB(Mode,COD,CBP,0,pic,bits,0);
					output_H263_CountBitsCoeff(pcoeff, Mode, CBP,bits,64);
				} else {
					output_H263_Count_sac_BitsMB(Mode,COD,CBP,0,pic,bits);
					output_H263_Count_sac_BitsCoeff(pcoeff, Mode, CBP,bits,64);
				}
			} else {
				for (k=0;k<6;k++) output_H263_Quant_blk(coeff,qcoeff,QP,Mode,k);
				output_H263_Scan(qcoeff,0);
				CBP = output_H263_FindCBP(qcoeff,Mode,64);
				if (!output_H263_syntax_arith_coding) {
					output_H263_CountBitsMB(Mode,COD,CBP,0,pic,bits,0);
					output_H263_CountBitsCoeff(qcoeff, Mode, CBP,bits,64);
				} else {
					output_H263_Count_sac_BitsMB(Mode,COD,CBP,0,pic,bits);
					output_H263_Count_sac_BitsCoeff(qcoeff, Mode, CBP,bits,64);
				}
				output_H263_DeScan(qcoeff,0);
				for (k=0;k<6;k++) output_H263_Dequant(qcoeff, rcoeff, QP, Mode,k);
				output_H263_MB_Decode(rcoeff, data);
			}
			output_H263_Clip(data);
			output_H263_ReconImage(i,j,data,recon);

			/* every block is coded */
			output_H263_coded_map[j+1][i+1] = 2;
			/* Initialize the intra_mb_refresh rate randomly after I-frame */
			if (output_H263_intra_mb_refresh) output_H263_intra_refresh[j+1][i+1] = (rand()%(output_H263_intra_mb_refresh+1));
			free(coeff);
		}
	}
	pic->QP_mean = (float)QP;

	if (output_H263_deblocking_filter) {
		output_H263_EdgeFilter(recon,pic);
	}

	if (output_H263_advanced_intra_coding) {
		free(store_coeff);
		free(store_rcoeff);
		free(pcoeff);
	}
	free(qcoeff);
	free(data);
	free(rcoeff);
	return recon;
}

