/************************************************************************
 *
 *  intra_pred.c, Intra Prediction routines of tmn (TMN encoder)
 *  for Advanced Intra Coding Mode
 *
 *  Copyright (C) 1997  University of BC, Canada
 *
 *  Contacts:
 *  Michael Gallant                   <mikeg@ee.ubc.ca>
 *  Guy Cote                          <guyc@ee.ubc.ca>
 *  Berna Erol                        <bernae@ee.ubc.ca>
 *
 *  UBC Image Processing Laboratory   http://www.ee.ubc.ca/image
 *  2356 Main Mall                    tel.: +1 604 822 4051
 *  Vancouver BC Canada V6T1Z4        fax.: +1 604 822 5949
 *
 *
 ************************************************************************/

/* Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee
 * or royalty on an "as is" basis. The University of British Columbia
 * disclaims any and all warranties, whether express, implied, or
 * statuary, including any implied warranties or merchantability or of
 * fitness for a particular purpose.  In no event shall the
 * copyright-holder be liable for any incidental, punitive, or
 * consequential damages of any kind whatsoever arising from the use of
 * these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and
 * user's customers, employees, agents, transferees, successors, and
 * assigns.
 *
 * The University of British Columbia does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any
 * third-party patents.
 *
 * Commercial implementations of H.263, including shareware, are subject to
 * royalty fees to patent holders.  Many of these patents are general
 * enough such that they are unavoidable regardless of implementation
 * design.
 *
 */

#include "sim.h"
#include "../error.h"

/**********************************************************************
 *
 *	Name:           Choose_AC_DC_Pred
 *	Description:    Intra Prediction in Advanced Intra Coding
 *
 *	Input:	        qcoeff, store_coeff, position of MB
 *
 *	Side effects:   change qcoeff to predicted qcoeff
 *
 *      Return:         Intra_Mode (used for the scanning order)
 *
 *	Date:970717     Guy Cote <guyc@ee.ubc.ca>
 *
 ***********************************************************************/

int output_H263_Choose_Intra_Mode (int *pcoeff, int *store_coeff, int xpos, int ypos, int newgob)
{
	int E0[6], E1[8][6], E2[8][6], A[8][6], B[8][6];
	int i, j;
	int SAD0 = 0, SAD1 = 0, SAD2 = 0;
	int Intra_Mode;


	for (i = 0; i < 6; i++) {

		if (xpos == 0 && ypos == 0) {
			/* top left corner */
			((i == 2 || i == 3)) ? output_H263_fill_A (A, store_coeff, xpos, ypos, i - 2, i) : output_H263_fill_null (A, i);
			((i == 1 || i == 3)) ? output_H263_fill_B (B, store_coeff, xpos, ypos, i - 1, i) : output_H263_fill_null (B, i);

		} else {
			if (xpos == 0) {
				/* left edge of the picture */
				(i == 2 || i == 3) ? output_H263_fill_A (A, store_coeff, xpos, ypos, i - 2, i) :
					((i == 0 || i == 1) && !(newgob)) ? output_H263_fill_A (A, store_coeff, xpos, ypos - 1, i + 2, i) :
					((i == 4 || i == 5) && !(newgob)) ? output_H263_fill_A (A, store_coeff, xpos, ypos - 1, i, i) :
					output_H263_fill_null (A, i);

				(i == 1 || i == 3) ? output_H263_fill_B (B, store_coeff, xpos, ypos, i - 1, i) : output_H263_fill_null (B, i);


			} else {
				if (ypos == 0) {
					/* top border of picture */
					if (i == 2 || i == 3) output_H263_fill_A(A, store_coeff, xpos, ypos, i - 2, i);
					else output_H263_fill_null(A, i);
					if ((i == 4 || i == 5) && output_H263_coded_map[ypos + 1][xpos] == 2) {
						output_H263_fill_B(B, store_coeff, xpos - 1, ypos, i, i);
					} else {
						if (i == 1 || i == 3) {
							output_H263_fill_B (B, store_coeff, xpos, ypos, i - 1, i);
						} else {
							if ((i == 0 || i == 2) && output_H263_coded_map[ypos + 1][xpos] == 2) {
								output_H263_fill_B (B, store_coeff, xpos - 1, ypos, i + 1, i);
							} else {
								output_H263_fill_null (B, i);
							}
						}
					}
				} else {
					/* anywhere else in the picture, do no
					 * cross GOB boundary */
					if (i == 2 || i == 3) {
						output_H263_fill_A(A, store_coeff, xpos, ypos, i - 2, i);
					} else {
						if ((i == 0 || i == 1) && !(newgob) && output_H263_coded_map[ypos][xpos + 1] == 2) {
							output_H263_fill_A(A, store_coeff, xpos, ypos - 1, i + 2, i);
						} else {
							if ((i == 4 || i == 5) && !(newgob) && output_H263_coded_map[ypos][xpos + 1] == 2) {
								output_H263_fill_A(A, store_coeff, xpos, ypos - 1, i, i);
							} else {
								output_H263_fill_null(A, i);
							}
						}
					}
					if ((i == 4 || i == 5) && output_H263_coded_map[ypos + 1][xpos] == 2) {
						output_H263_fill_B(B, store_coeff, xpos - 1, ypos, i, i);
					} else {
						if (i == 1 || i == 3) {
							output_H263_fill_B(B, store_coeff, xpos, ypos, i - 1, i);
						} else {
							if ((i == 0 || i == 2) && output_H263_coded_map[ypos + 1][xpos] == 2) {
								output_H263_fill_B(B, store_coeff, xpos - 1, ypos, i + 1, i);
							} else {
								output_H263_fill_null(B, i);
							}
						}
					}
				}
			}
		}

		/* DC prediction */

		E0[i] = store_coeff[(xpos + ypos * output_H263_pels / MB_SIZE) * 384 + i * 64] - (A[0][i] + B[0][i] + 1) / 2;

		for (j = 0; j < 8; j++) {
			/* DC/AC prediction from above (block A) */
			E1[j][i] = store_coeff[(xpos + ypos * output_H263_pels / MB_SIZE) * 384 + j + i * 64] - A[j][i];
			/* DC/AC prediction from left (block B) */
			E2[j][i] = store_coeff[(xpos + ypos * output_H263_pels / MB_SIZE) * 384 + j * 8 + i * 64] - B[j][i];
		}
		/* Choose Intra Mode based on SAD of the prediction error */
		SAD0 += E0[i];
		SAD1 += E1[0][i];
		SAD2 += E2[0][i];
		for (j = 1; j < 8; j++) {
			SAD0 += 32 * abs (store_coeff[(xpos + ypos * output_H263_pels / MB_SIZE) * 384 + j + i * 64])
				+ 32 * abs (store_coeff[(xpos + ypos * output_H263_pels / MB_SIZE) * 384 + j * 8 + i * 64]);
			SAD1 += 32 * abs (E1[j][i]) + 32 * abs (store_coeff[(xpos + ypos * output_H263_pels / MB_SIZE) * 384 + j * 8 + i * 64]);
			SAD2 += 32 * abs (store_coeff[(xpos + ypos *  output_H263_pels / MB_SIZE) * 384 + j + i * 64]) + 32 * abs (E2[j][i]);
		}
	}

	if (SAD0 <= SAD1) {
		if (SAD0 <= SAD2) Intra_Mode = INTRA_MODE_DC;
		else Intra_Mode = INTRA_MODE_HORI_AC;
	} else if (SAD1 <= SAD2) Intra_Mode = INTRA_MODE_VERT_AC;
	else Intra_Mode = INTRA_MODE_HORI_AC;

	return Intra_Mode;
}

void output_H263_Intra_AC_DC_Encode (int *pcoeff, int *store_rcoeff, int Intra_Mode, int xpos, int ypos, int newgob, int i)
{

	int A[8][6], B[8][6];
	int j;


	if (xpos == 0 && ypos == 0) {
		/* top left corner */
		(i == 2 || i == 3) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos, i - 2, i) : output_H263_fill_null (A, i);
		(i == 1 || i == 3) ? output_H263_fill_B (B, store_rcoeff, xpos, ypos, i - 1, i) : output_H263_fill_null (B, i);
	} else {
		/* left border of picture */
		if (xpos == 0)
		{                           /* left edge of the picture */
			(i == 2 || i == 3) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos, i - 2, i) :
				((i == 0 || i == 1) && !(newgob)) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos - 1, i + 2, i) :
				((i == 4 || i == 5) && !(newgob)) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos - 1, i, i) : output_H263_fill_null (A, i);
			(i == 1 || i == 3) ? output_H263_fill_B (B, store_rcoeff, xpos, ypos, i - 1, i) : output_H263_fill_null (B, i);

		} else {
			if (ypos == 0) {
				/* top border of picture */
				(i == 2 || i == 3) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos, i - 2, i) : output_H263_fill_null (A, i);
				(i == 4 || i == 5) ? output_H263_fill_B (B, store_rcoeff, xpos - 1, ypos, i, i) :
					(i == 1 || i == 3) ? output_H263_fill_B (B, store_rcoeff, xpos, ypos, i - 1, i) :
					output_H263_fill_B (B, store_rcoeff, xpos - 1, ypos, i + 1, i);
			} else {
				/* anywhere else in the picture, do no
				 * cross GOB boundary */
				(i == 2 || i == 3) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos, i - 2, i) :
					((i == 0 || i == 1) && !(newgob)) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos - 1, i + 2, i) :
					((i == 4 || i == 5) && !(newgob)) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos - 1, i, i) : output_H263_fill_null (A, i);

				(i == 4 || i == 5) ? output_H263_fill_B (B, store_rcoeff, xpos - 1, ypos, i, i) :
					(i == 1 || i == 3) ? output_H263_fill_B (B, store_rcoeff, xpos, ypos, i - 1, i) :
					output_H263_fill_B (B, store_rcoeff, xpos - 1, ypos, i + 1, i);

			}
		}
	}


	/* replace the qcoeff with the predicted values pcoeff */

	switch (Intra_Mode) {
	case INTRA_MODE_DC:
		/* It is OK to use 1024 to check the DC direction since the DC *
		 * reconstructed coefficient will always be odd                */
		pcoeff[i * 64] -=
			(A[0][i] == 1024 && B[0][i] == 1024) ? 1024 :
				(A[0][i] == 1024) ? B[0][i] :
					(B[0][i] == 1024) ? A[0][i] : ((A[0][i] + B[0][i]) / 2);

		break;
	case INTRA_MODE_VERT_AC:
		for (j = 0; j < 8; j++) {
			pcoeff[j + i * 64] -= A[j][i];
		}
		break;
	case INTRA_MODE_HORI_AC:
		for (j = 0; j < 8; j++) {
			pcoeff[j * 8 + i * 64] -= B[j][i];
		}
		break;
	default:
		ERROR(INTMSG("Error in Prediction in Advanced Intra Coding"));
		END(1);
		break;
	}

	return;
}

/**********************************************************************
 *
 *	Name:           fill_null, fill_A and fill_B
 *	Description:    Fill values in predictor coefficients
 *                      Functions used in advanced intra coding mode
 *
 *	Input:	        predictor qcoefficients
 *	Side effects:
 *
 *      Return:
 *
 *	Date:970717     Guy Cote <guyc@ee.ubc.ca>
 *
 ***********************************************************************/


void output_H263_fill_null (int pred[][6], int i)
{
	int j;

	pred[0][i] = 1024;

	for (j = 1; j < 8; j++) {
		pred[j][i] = 0;
	}
}

void output_H263_fill_A (int pred[][6], int *store_rcoeff, int xpos, int ypos, int block, int i)
{
	/* Fill first row of block at MB xpos, ypos, in pred[][i] */
	int j;

	for (j = 0; j < 8; j++) {
		pred[j][i] = *(store_rcoeff + (ypos * output_H263_pels / MB_SIZE + xpos) * 384 + block * 64 + j);
	}
}

void output_H263_fill_B (int pred[][6], int *store_rcoeff, int xpos, int ypos, int block, int i)
{
	/* Fill first column of block at MB xpos, ypos, in pred[][i] */
	int j;

	for (j = 0; j < 8; j++) {
		pred[j][i] = *(store_rcoeff + (ypos * output_H263_pels / MB_SIZE + xpos) * 384 + block * 64 + j * 8);
	}
}


/**********************************************************************
 *
 *	Name:           Intra_AC_DC_Decode
 *	Description:    Intra Prediction in Advanced Intra Coding
 *
 *	Input:	        store_rcoeff, Intra_AC_DC, position of MB, store_QP
 *
 *	Side effects:   change qcoeff to predicted qcoeff
 *
 *      Return:
 *
 *	Date:970717     Guy Cote <guyc@ee.ubc.ca>
 *
 ***********************************************************************/


void output_H263_Intra_AC_DC_Decode (int *rcoeff, int *store_rcoeff, int INTRA_AC_DC, int xpos, int ypos, int newgob, int i)
{

	int A[8][6], B[8][6];
	int j, k, tempDC;


	if (xpos == 0 && ypos == 0) {
		/* top left corner */
		(i == 2 || i == 3) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos, i - 2, i) : output_H263_fill_null (A, i);
		(i == 1 || i == 3) ? output_H263_fill_B (B, store_rcoeff, xpos, ypos, i - 1, i) : output_H263_fill_null (B, i);
	} else
	{                             /* left border of picture */
		if (xpos == 0)
		{                           /* left edge of the picture */
			(i == 2 || i == 3) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos, i - 2, i) :
				((i == 0 || i == 1) && !(newgob)) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos - 1, i + 2, i) :
				((i == 4 || i == 5) && !(newgob)) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos - 1, i, i) : output_H263_fill_null (A, i);
			(i == 1 || i == 3) ? output_H263_fill_B (B, store_rcoeff, xpos, ypos, i - 1, i) : output_H263_fill_null (B, i);

		} else {
			if (ypos == 0)
			{                         /* top border of picture */
				(i == 2 || i == 3) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos, i - 2, i) : output_H263_fill_null (A, i);
				(i == 4 || i == 5) ? output_H263_fill_B (B, store_rcoeff, xpos - 1, ypos, i, i) :
					(i == 1 || i == 3) ? output_H263_fill_B (B, store_rcoeff, xpos, ypos, i - 1, i) :
					output_H263_fill_B (B, store_rcoeff, xpos - 1, ypos, i + 1, i);
			} else {
				/* anywhere else in the picture, do no
				 * cross GOB boundary */
				(i == 2 || i == 3) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos, i - 2, i) :
					((i == 0 || i == 1) && !(newgob)) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos - 1, i + 2, i) :
					((i == 4 || i == 5) && !(newgob)) ? output_H263_fill_A (A, store_rcoeff, xpos, ypos - 1, i, i) : output_H263_fill_null (A, i);

				(i == 4 || i == 5) ? output_H263_fill_B (B, store_rcoeff, xpos - 1, ypos, i, i) :
					(i == 1 || i == 3) ? output_H263_fill_B (B, store_rcoeff, xpos, ypos, i - 1, i) :
					output_H263_fill_B (B, store_rcoeff, xpos - 1, ypos, i + 1, i);
			}
		}
	}

	switch (INTRA_AC_DC) {
	case INTRA_MODE_DC:
		/* It is OK to use 1024 to check the DC direction since the DC *
		 * reconstructed coefficient will always be odd                */
		tempDC = rcoeff[i * 64]
			+ ((A[0][i] == 1024 && B[0][i] == 1024) ? 1024 :
			   (A[0][i] == 1024) ? B[0][i] :
			   (B[0][i] == 1024) ? A[0][i] : (A[0][i] + B[0][i]) / 2);
		for (j = 0; j < 8; j++)
			for (k = 0; k < 8; k++)
				rcoeff[i * 64 + j * 8 + k] = output_H263_clipAC (rcoeff[i * 64 + j * 8 + k]);
		rcoeff[i * 64] = output_H263_oddifyclipDC (tempDC);
		break;

	case INTRA_MODE_VERT_AC:
		tempDC = rcoeff[i * 64] + A[0][i];
		for (j = 1; j < 8; j++) {
			rcoeff[i * 64 + j] = output_H263_clipAC (rcoeff[i * 64 + j] + A[j][i]);
		}
		for (j = 1; j < 8; j++)
			for (k = 0; k < 8; k++)
				rcoeff[i * 64 + j * 8 + k] = output_H263_clipAC (rcoeff[i * 64 + j * 8 + k]);
		rcoeff[i * 64] = output_H263_oddifyclipDC (tempDC);
		break;
	case INTRA_MODE_HORI_AC:
		tempDC = rcoeff[i * 64] + B[0][i];
		for (j = 1; j < 8; j++) rcoeff[i * 64 + j * 8] = output_H263_clipAC(rcoeff[i * 64 + j * 8] + B[j][i]);
		for (j = 0; j < 8; j++) for (k = 1; k < 8; k++) rcoeff[i * 64 + j * 8 + k] = output_H263_clipAC(rcoeff[i * 64 + j * 8 + k]);
		rcoeff[i * 64] = output_H263_oddifyclipDC(tempDC);
		break;
	default:
		ERROR(INTMSG("Error in Prediction in Advanced Intra Coding"));
		END(1);
		break;
	}

	/* store the  reconstructed DCT coefficients */
	memcpy((void *) (store_rcoeff + i * 64 + (xpos + ypos * output_H263_pels / MB_SIZE) * 384), (void *) (rcoeff + i * 64), sizeof (int) * 64);


	return;
}

/**********************************************************************
 *
 *	Name:           oddifyclipDC, clipAC and clipDC
 *	Description:    Fill values in predictor coefficients
 *                      Functions used in advanced intra coding mode
 *
 *	Input:	        predictor qcoefficients
 *	Side effects:
 *
 *      Return:
 *
 *	Date:970717     Guy Cote <guyc@ee.ubc.ca>
 *
 ***********************************************************************/


int output_H263_oddifyclipDC (int x)
{
	int result;

	result = output_H263_clipDC(x + ((x&0x1)^0x1));
	return result;
}

int output_H263_clipAC (int x)
{
	int clipped;

	if (x > 2047) clipped = 2047;
	else if (x < -2048) clipped = -2048;
	else clipped = x;
	return clipped;

}
int output_H263_clipDC (int x)
{
	int clipped;
	if (x > 2047) clipped = 2047;
	else if (x < 0) clipped = 0;
	else clipped = x;
	return clipped;
}
