/************************************************************************
 *
 *  global.h, global variables for tmndecode (H.263 decoder)
 *  Copyright (C) 1995, 1996  Telenor R&D, Norway
 *
 *  Contacts:
 *  Robert Danielsen                  <Robert.Danielsen@nta.no>
 *
 *  Telenor Research and Development  http://www.nta.no/brukere/DVC/
 *  P.O.Box 83                        tel.:   +47 63 84 84 00
 *  N-2007 Kjeller, Norway            fax.:   +47 63 81 00 76
 *
 *  Copyright (C) 1997  University of BC, Canada
 *  Modified by: Michael Gallant <mikeg@ee.ubc.ca>
 *               Guy Cote <guyc@ee.ubc.ca>
 *               Berna Erol <bernae@ee.ubc.ca>
 *
 *  Contacts:
 *  Michael Gallant                   <mikeg@ee.ubc.ca>
 *
 *  UBC Image Processing Laboratory   http://www.ee.ubc.ca/image
 *  2356 Main Mall                    tel.: +1 604 822 4051
 *  Vancouver BC Canada V6T1Z4        fax.: +1 604 822 5949
 *
 ************************************************************************/

/* Disclaimer of Warranty
 * 
 * These software programs are available to the user without any license fee
 * or royalty on an "as is" basis. The University of British Columbia
 * disclaims any and all warranties, whether express, implied, or
 * statuary, including any implied warranties or merchantability or of
 * fitness for a particular purpose.  In no event shall the
 * copyright-holder be liable for any incidental, punitive, or
 * consequential damages of any kind whatsoever arising from the use of
 * these programs.
 * 
 * This disclaimer of warranty extends to the user of these programs and
 * user's customers, employees, agents, transferees, successors, and
 * assigns.
 * 
 * The University of British Columbia does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any
 * third-party patents.
 * 
 * Commercial implementations of H.263, including shareware, are subject to
 * royalty fees to patent holders.  Many of these patents are general
 * enough such that they are unavoidable regardless of implementation
 * design.
 * 
 */



/* based on mpeg2decode, (C) 1994, MPEG Software Simulation Group and
 * mpeg2play, (C) 1994 Stefan Eckart <stefan@lis.e-technik.tu-muenchen.de>
 * 
 */


/* choose between declaration (GLOBAL undefined) and definition (GLOBAL
 * defined) GLOBAL is defined in exactly one file (tmndec.c) */

#include <stdio.h>

#ifndef GLOBAL
#define EXTERN extern
#define INIT(x)
#else
#define EXTERN
#define INIT(x) = x
#endif

/* Some macros */
#define mmax(a, b)        ((a) > (b) ? (a) : (b))
#define mmin(a, b)        ((a) < (b) ? (a) : (b))
#define mnint(a)        ((a) < 0 ? (int)(a - 0.5) : (int)(a + 0.5))
#define sign(a)         ((a) < 0 ? -1 : 1)

typedef struct {
  int val, len;
} VLCtab;
typedef struct {
	char run, level, len;
} DCTtab;
typedef struct {
	int val, len;
} VLCtabI;
typedef struct {
	int val, run, sign;
} RunCoef;
typedef struct {
	int code, level, last, run, sign, extended_level;
} CodeCoeff;

/* prototypes of global functions */

/* getbits.c */
void read_H263_initbits(void);
void read_H263_fillbfr(void);
unsigned int read_H263_showbits(int n);
unsigned int read_H263_getbits1(void);
void read_H263_flushbits(int n);
unsigned int read_H263_getbits(int n);

/* getblk.c */
void read_H263_getblock(int comp, int mode, int INTRA_AC_DC, int Mode);
void read_H263_get_sac_block(int comp, int mode, int INTRA_AC_DC, int Mode);

/* gethdr.c */
int read_H263_getheader(void);
void read_H263_startcode(void);
void read_H263_initoptions(void);
void read_H263_getgobheader(void);
/* getpic.c */
void read_H263_getpicture(int *framenum, int gob);
void read_H263_putlast(int framenum);
void read_H263_PostFilter(unsigned char *lum, unsigned char *Cb, unsigned char *Cr, int width, int height);

/* scalability funcitons */
void read_H263_UpsampleReferenceLayerPicture(void);
void read_H263_UpsampleComponent(unsigned char *enhanced, unsigned char *base, int horiz, int vert);
/* advanced intra coding functions */
void read_H263_Intra_AC_DC_Decode(short *store_qcoeff, int INTRA_AC_DC, int MBA, int xpos, int ypos, int comp, int newgob);
void read_H263_fill_null(int pred[]);
void read_H263_fill_A(int pred[], short *store_qcoeff, int xpos, int ypos, int block);
void read_H263_fill_B(int pred[], short *store_qcoeff, int xpos, int ypos, int block);
int read_H263_oddifyclipDC(int x);
int read_H263_clipAC(int x);
int read_H263_clipDC(int x);
void read_H263_interpolate_image(unsigned char *in, unsigned char *out, int width, int height);

/* getvlc.c */
int read_H263_getTMNMV(void);
int read_H263_getRVLC (void);
int read_H263_getMCBPC(void);
int read_H263_getMODB(void);
int read_H263_getMCBPCintra(void);
int read_H263_getCBPY(void);
int read_H263_getMBTYPE(int *true_B_cbp, int *true_B_quant);
int read_H263_getscalabilityCBPC();

/* idct.c */
void read_H263_idct(short *block);
void read_H263_init_idct(void);

/* idctref.c */
void read_H263_init_idctref(void);
void read_H263_idctref(short *block);

/* recon.c */
void read_H263_reconstruct(int bx, int by, int P, int bdx, int bdy, int MODB, int newgob);
void read_H263_reconstruct_true_B(int bx, int by, int true_B_prediction_type);
void read_H263_reconstruct_ei_ep(int bx, int by, int ei_ep_prediction_type);

/* sac.c */
int read_H263_decode_a_symbol(int cumul_freq[]);
void read_H263_decoder_reset();

/* ring_buf.c */
int read_H263_get_reference_picture(void);
void read_H263_store_picture(int quality);

/* zig-zag scan */
EXTERN unsigned char read_H263_zig_zag_scan[64]
#ifdef GLOBAL
=
{
 0, 1, 8, 16, 9, 2, 3, 10, 17, 24, 32, 25, 18, 11, 4, 5,
 12, 19, 26, 33, 40, 48, 41, 34, 27, 20, 13, 6, 7, 14, 21, 28,
 35, 42, 49, 56, 57, 50, 43, 36, 29, 22, 15, 23, 30, 37, 44, 51,
 58, 59, 52, 45, 38, 31, 39, 46, 53, 60, 61, 54, 47, 55, 62, 63
}
#endif
;

EXTERN int read_H263_lines[]
#ifdef GLOBAL
= {-1,128,176,352,704,1408,-1}
#endif
;

EXTERN int read_H263_pels[7]
#ifdef GLOBAL
= {-1,96,144,288,576,1152,-1}
#endif
;

/* other scan orders needed in advanced intra coding mode */
EXTERN unsigned char read_H263_alternate_horizontal_scan[64]
#ifdef GLOBAL
=
{
 0, 1, 2, 3, 8, 9, 16, 17, 10, 11, 4, 5, 6, 7, 15, 14,
 13, 12, 19, 18, 24, 25, 32, 33, 26, 27, 20, 21, 22, 23, 28, 29,
 30, 31, 34, 35, 40, 41, 48, 49, 42, 43, 36, 37, 38, 39, 44, 45,
 46, 47, 50, 51, 56, 57, 58, 59, 52, 53, 54, 55, 60, 61, 62, 63
}
#endif
;
EXTERN unsigned char read_H263_alternate_vertical_scan[64]
#ifdef GLOBAL
=
{
 0, 8, 16, 24, 1, 9, 2, 10, 17, 25, 32, 40, 48, 56, 57, 49,
 41, 33, 26, 18, 3, 11, 4, 12, 19, 27, 34, 42, 50, 58, 35, 43,
 51, 59, 20, 28, 5, 13, 6, 14, 21, 29, 36, 44, 52, 60, 37, 45,
 53, 61, 22, 30, 7, 15, 23, 31, 38, 46, 54, 62, 39, 47, 55, 63
}
#endif
;

/* color space conversion coefficients
 * 
 * entries are {crv,cbu,cgu,cgv}
 * 
 * crv=(255/224)*65536*(1-cr)/0.5 cbu=(255/224)*65536*(1-cb)/0.5
 * cgu=(255/224)*65536*(cb/cg)*(1-cb)/0.5
 * cgv=(255/224)*65536*(cr/cg)*(1-cr)/0.5
 * 
 * where Y=cr*R+cg*G+cb*B(cr+cg+cb=1) */

EXTERN int read_H263_convmat[8][4]
#ifdef GLOBAL
=
{
 {117504, 138453, 13954, 34903}, /* no sequence_display_extension */
 {117504, 138453, 13954, 34903}, /* ITU-R Rec. 709(1990) */
 {104597, 132201, 25675, 53279}, /* unspecified */
 {104597, 132201, 25675, 53279}, /* reserved */
 {104448, 132798, 24759, 53109}, /* FCC */
 {104597, 132201, 25675, 53279}, /* ITU-R Rec. 624-4 System B, G */
 {104597, 132201, 25675, 53279}, /* SMPTE 170M */
 {117579, 136230, 16907, 35559}/* SMPTE 240M(1987) */
}
#endif
;

EXTERN int read_H263_gotframe;
EXTERN int read_H263_quiet;
EXTERN unsigned char *read_H263_prev_frame[3], *read_H263_current_frame[3], *read_H263_bframe[3];
EXTERN unsigned char *read_H263_tmp_f[3], *read_H263_tmp_b[3];
EXTERN unsigned char *read_H263_edgeframe[3], *read_H263_edgeframeorig[3];
EXTERN unsigned char *read_H263_nextedgeframe[3], *read_H263_nextedgeframeorig[3];
EXTERN unsigned char *read_H263_enhancementedgeframe[3], *read_H263_enhacnementedgeframeorig[3];
EXTERN unsigned char *read_H263_next_I_P_frame[3], *read_H263_prev_I_P_frame[3], *read_H263_buffered_frame[3];
EXTERN int read_H263_MV[2][6][MBR + 1][MBC + 2];
EXTERN int read_H263_true_B_direct_mode_MV[2][5][MBR + 1][MBC + 2];
EXTERN int read_H263_modemap[MBR + 1][MBC + 2];
EXTERN int read_H263_anchorframemodemap[MBR + 1][MBC + 2];
EXTERN int read_H263_predictionmap[MBR + 1][MBC + 2];
EXTERN unsigned char *read_H263_clp;
EXTERN int read_H263_horizontal_size, read_H263_vertical_size, read_H263_mb_width, read_H263_mb_height;
EXTERN int read_H263_coded_picture_width, read_H263_coded_picture_height;
EXTERN int read_H263_ref_coded_picture_width, read_H263_ref_coded_picture_height;
EXTERN int read_H263_chrom_width, read_H263_chrom_height, read_H263_blk_cnt;
EXTERN int read_H263_ref_chrom_width, read_H263_ref_chrom_height;
EXTERN int read_H263_pict_type, read_H263_newgob, read_H263_gfid, read_H263_pgfid;
EXTERN int read_H263_mv_outside_frame;
EXTERN int read_H263_syntax_arith_coding;
EXTERN int read_H263_adv_pred_mode;
EXTERN int read_H263_overlapping_MC;
EXTERN int read_H263_use_4mv;
EXTERN int read_H263_pb_frame;
EXTERN int read_H263_true_B_frame;
EXTERN int read_H263_scalability_mode;
EXTERN int read_H263_long_vectors;
EXTERN int read_H263_fault, read_H263_expand;
EXTERN int read_H263_verbose;
EXTERN int read_H263_refidct;
EXTERN int read_H263_matrix_coefficients;
EXTERN int read_H263_temp_ref, read_H263_prev_non_disposable_temp_ref, read_H263_next_non_disposable_temp_ref;
EXTERN int read_H263_pic_quant, read_H263_quant, read_H263_source_format, read_H263_base_source_format;
EXTERN int read_H263_UFEP;
EXTERN int read_H263_prev_mv_outside_frame, read_H263_prev_sac, read_H263_prev_adv_pred, read_H263_prev_aic, read_H263_prev_df;
EXTERN int read_H263_prev_slice_struct, read_H263_prev_rps, read_H263_prev_isd, read_H263_prev_aivlc, read_H263_prev_mq;
EXTERN int read_H263_prev_4mv, read_H263_prev_long_vectors, read_H263_prev_obmc;

/* Following variables are used in H.263+ bitstream decoding */
EXTERN int read_H263_plus_type;      /* indicates if extended PTYPE exists or not. */
EXTERN int read_H263_optional_custom_PCF;
EXTERN int read_H263_advanced_intra_coding;
EXTERN int read_H263_deblocking_filter_mode;
EXTERN int read_H263_slice_structured_mode;
EXTERN int read_H263_reference_picture_selection_mode;
EXTERN int read_H263_independently_segmented_decoding_mode;
EXTERN int read_H263_alternative_inter_VLC_mode;
EXTERN int read_H263_modified_quantization_mode;
EXTERN int read_H263_reduced_resolution_update_mode;
EXTERN int read_H263_reference_picture_resampling_mode;
EXTERN int read_H263_rtype;
EXTERN int read_H263_post_filter;
EXTERN int read_H263_unlimited_unrestricted_motion_vectors;
EXTERN int read_H263_concealment;

/* Scalability globals */
EXTERN unsigned char *read_H263_prev_enhancement_frame[MAX_LAYERS][3], *read_H263_current_enhancement_frame[MAX_LAYERS][3];
EXTERN unsigned char *read_H263_curr_reference_frame[3], *read_H263_upsampled_reference_frame[3];
EXTERN unsigned char *read_H263_enhance_edgeframeorig[MAX_LAYERS][3], *read_H263_enhance_edgeframe[MAX_LAYERS][3];
EXTERN unsigned char *read_H263_tmp_enhance_fwd[MAX_LAYERS][3], *read_H263_tmp_enhance_up[MAX_LAYERS][3];
EXTERN int read_H263_enhancement_layer_init[MAX_LAYERS];
EXTERN int read_H263_enhance_pict;
 
/* custom format variables */
EXTERN int read_H263_CP_PAR_code;
EXTERN int read_H263_CP_picture_width_indication;
EXTERN int read_H263_CP_picture_width;
EXTERN int read_H263_CP_picture_height_indication;
EXTERN int read_H263_CP_picture_height;
EXTERN int read_H263_PAR_width;
EXTERN int read_H263_PAR_height;

EXTERN int read_H263_CP_clock_frequency;
EXTERN int read_H263_SSS_rectangular_slices;
EXTERN int read_H263_SSS_arbitary_slice_ordering;
EXTERN int read_H263_enhancement_layer_num;
EXTERN int read_H263_reference_layer_number;
EXTERN int read_H263_MF_of_reference_picture_selection;
EXTERN int read_H263_TRI, read_H263_TRPI;
EXTERN int read_H263_temporal_reference_for_prediction;

EXTERN int read_H263_trd, read_H263_trb, read_H263_bscan, read_H263_bquant, read_H263_true_b_trb;
EXTERN int read_H263_bscan_tab[]
#ifdef GLOBAL
= {2, 4, 6, 8}
#endif
;
EXTERN int read_H263_bquant_tab[]
#ifdef GLOBAL
= {5, 6, 7, 8}
#endif
;

EXTERN int read_H263_OM[5][8][8]
#ifdef GLOBAL
= {
 {
  {4, 5, 5, 5, 5, 5, 5, 4},
  {5, 5, 5, 5, 5, 5, 5, 5},
  {5, 5, 6, 6, 6, 6, 5, 5},
  {5, 5, 6, 6, 6, 6, 5, 5},
  {5, 5, 6, 6, 6, 6, 5, 5},
  {5, 5, 6, 6, 6, 6, 5, 5},
  {5, 5, 5, 5, 5, 5, 5, 5},
  {4, 5, 5, 5, 5, 5, 5, 4},
 }, {
  {2, 2, 2, 2, 2, 2, 2, 2},
  {1, 1, 2, 2, 2, 2, 1, 1},
  {1, 1, 1, 1, 1, 1, 1, 1},
  {1, 1, 1, 1, 1, 1, 1, 1},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
 }, {
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {1, 1, 1, 1, 1, 1, 1, 1},
  {1, 1, 1, 1, 1, 1, 1, 1},
  {1, 1, 2, 2, 2, 2, 1, 1},
  {2, 2, 2, 2, 2, 2, 2, 2},
 }, {
  {0, 0, 0, 0, 1, 1, 1, 2},
  {0, 0, 0, 0, 1, 1, 2, 2},
  {0, 0, 0, 0, 1, 1, 2, 2},
  {0, 0, 0, 0, 1, 1, 2, 2},
  {0, 0, 0, 0, 1, 1, 2, 2},
  {0, 0, 0, 0, 1, 1, 2, 2},
  {0, 0, 0, 0, 1, 1, 2, 2},
  {0, 0, 0, 0, 1, 1, 1, 2},
 }, {
  {2, 1, 1, 1, 0, 0, 0, 0},
  {2, 2, 1, 1, 0, 0, 0, 0},
  {2, 2, 1, 1, 0, 0, 0, 0},
  {2, 2, 1, 1, 0, 0, 0, 0},
  {2, 2, 1, 1, 0, 0, 0, 0},
  {2, 2, 1, 1, 0, 0, 0, 0},
  {2, 2, 1, 1, 0, 0, 0, 0},
  {2, 1, 1, 1, 0, 0, 0, 0},
}}
#endif
;

EXTERN int read_H263_roundtab[16]
#ifdef GLOBAL
= {0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2}
#endif
;

EXTERN int read_H263_MQ_chroma_QP_table[32]
#ifdef GLOBAL
 = {0, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 
14, 14, 14, 14, 14, 15, 15, 15, 15, 15}
#endif
;

EXTERN struct ld {
 /* bit input */
 int infile;
 unsigned char rdbfr[2051];
 unsigned char *rdptr;
 unsigned char inbfr[16];
 int incnt;
 int bitcnt;
 /* block data */
 short block[12][64];
} read_H263_base, *read_H263_ld;


/* For reference picture selection Mode(Annex N) */

#define RING_SIZE 8

EXTERN int  read_H263_ring_pointer;        /* pointer - to last picture */
EXTERN void *read_H263_ring_lum[RING_SIZE];  /* luminance */
EXTERN void *read_H263_ring_Cr[RING_SIZE];   /* chominance */
EXTERN void *read_H263_ring_Cb[RING_SIZE];
EXTERN int  read_H263_ring_temporal_reference[RING_SIZE];   /* TR values for the pictures */
EXTERN int  read_H263_ring_quality[RING_SIZE];   /* picture quality */
EXTERN int  read_H263_stop_decoder;      /* flag if we completely out of sync */

/* Back Channel Message Structure */
/* used with Annex N with Videomux channel */
EXTERN struct bcm 
{
 int present;      /* set to 1 if bcm is present */
 int BT;        /* Back-channel message type (2 bits) */
 int URF;        /* Unreliable flag (1 bit) */
 int TR;        /* Temporal reference (10 bits) */
 int ELNUMI;      /* Enhancement layer number indication (1 bit) */
 int ELNUM;       /* Enhancement layer number (4 bits) */
 int BCPM;       /* Indicates that BSBI is present (1 bit) */
 int BSPI;       /* Back-channel sub bitstream indicator (2 bits) */
 int BEPB1;       /* Back-channel emulation prevention bit 1 (1 bit set to 1 */
 int GN_MBA;      /* GOB number / Macroblock address (variable length) */
 int BEPB2;       /* Back-channel emulation prevention bit 2 (1 bit set to 1 */
 int RTR;        /* Requested temporal reference (10 bits) */
 int BSTUF;       /* stuffing (variable length) */
} read_H263_BCM;

